from typing import Iterable, List

import geojson
import numpy as np
import scipy.ndimage
import xtgeo

MISSING_DEPENDENCIES = False
try:
    import shapely.geometry

    # Use _contour from MPL. Under the hood, this is the same functionality pyplot is
    # using. Direct use is implemented (instead of pyplot.contour directly) to avoid the
    # overhead related to figure/axis creation in MPL.
    from matplotlib import __version__ as mpl_ver
    from matplotlib import _contour
except ImportError:
    MISSING_DEPENDENCIES = True


def plume_polygons(
    surfaces: List[xtgeo.RegularSurface],
    threshold: float,
    smoothing: float = 10.0,
    simplify_factor: float = 1.2,
) -> geojson.FeatureCollection:
    if MISSING_DEPENDENCIES:
        return geojson.FeatureCollection(features=[])
    plume_count = truncate_surfaces(surfaces, threshold, smoothing)
    p_levels = [0.1]
    if len(surfaces) > 2:
        p_levels.append(0.5)
    if len(surfaces) > 1:
        p_levels.append(0.9)
    levels = [lvl * len(surfaces) for lvl in p_levels]
    contours = _extract_contours(plume_count, surfaces[0], simplify_factor, levels)
    return geojson.FeatureCollection(
        features=[
            geojson.Feature(
                id=f"P{level * 100}",
                geometry=geojson.LineString(poly),
            )
            for level, polys in zip(p_levels, contours)
            for poly in polys
        ]
    )


def truncate_surfaces(
    surfaces: List[xtgeo.RegularSurface], threshold: float, smoothing: float
) -> np.ndarray:
    binary = [
        (
            np.where(np.isnan(s.values) | s.values.mask, 0.0, s.values) > threshold
        ).astype(float)
        for s in surfaces
    ]
    count = sum(binary)
    return scipy.ndimage.gaussian_filter(count, sigma=smoothing, mode="nearest")


def _extract_contours(
    surface: np.ndarray,
    ref_surface: xtgeo.RegularSurface,
    simplify_factor: float,
    levels: List[float],
) -> List:
    x = ref_surface.xori + np.arange(0, ref_surface.ncol) * ref_surface.xinc
    y = ref_surface.yori + np.arange(0, ref_surface.nrow) * ref_surface.yinc
    res = np.mean([abs(x[1] - x[0]), abs(y[1] - y[0])])
    simplify_dist = simplify_factor * res
    return _find_all_contours(x, y, surface, levels, simplify_dist)


def _find_all_contours(
    x_lin: np.ndarray,
    y_lin: np.ndarray,
    z_values: np.ndarray,
    levels: List[float],
    simplify_dist: float,
) -> List[List[List[List[float]]]]:
    x_mesh, y_mesh = np.meshgrid(x_lin, y_lin, indexing="ij")
    polys = [
        [
            _simplify(poly, simplify_dist)
            for poly in _find_contours(x_mesh, y_mesh, z_values >= level)
        ]
        for level in levels
    ]
    return polys


def _find_contours(
    x_mesh: np.ndarray,
    y_mesh: np.ndarray,
    z_values: np.ndarray,
) -> Iterable[np.ndarray]:
    contour_output = _contour.QuadContourGenerator(
        x_mesh, y_mesh, z_values, np.zeros_like(z_values, dtype=bool), False, 0
    ).create_contour(0.5)
    if int(mpl_ver[0]) >= 3 and int(mpl_ver[2]) >= 5:
        contour_output = contour_output[0]
    return contour_output


def _simplify(poly: np.ndarray, simplify_dist: float) -> List[List[float]]:
    simplified = shapely.geometry.LineString(poly).simplify(simplify_dist)
    return np.array(simplified.coords).tolist()



from typing import Callable, Dict, List, Tuple

from dash import Dash, html
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.deprecation_decorators import deprecated_plugin

from ._business_logic import WellCompletionsDataModel
from ._callbacks import plugin_callbacks
from ._layout import layout_tour_steps, main_layout


@deprecated_plugin(
    "This plugin has been replaced by the `WellCompletion` plugin (without "
    "the `s`) which is based on the `wellcompletionsdata` export from `ecl2df`. "
    "The new plugin is faster and has more functionality. "
)
class WellCompletions(WebvizPluginABC):
    # pylint: disable=line-too-long
    """Visualizes well completions data per well coming from export of the Eclipse COMPDAT output. \
    Data is grouped per well and zone and can be filtered accoring to flexible well categories.

    ---

    * **`ensembles`:** Which ensembles in `shared_settings` to visualize.
    * **`compdat_file`:** `.csv` file with compdat data per realization
    * **`well_connection_status_file`:** `.parquet` file with well connection status per realization
    * **`zone_layer_mapping_file`:** `.lyr` file specifying the zone ➔ layer mapping
    * **`stratigraphy_file`:** `.json` file defining the stratigraphic levels
    * **`well_attributes_file`:** `.json` file with categorical well attributes
    * **`kh_unit`:** e.g. mD·m, will try to extract from eclipse files if defaulted
    * **`kh_decimal_places`:**

    ---
    The minimum requirement is to define `ensembles`.

    **COMPDAT input**

    `compdat_file` is a path to a file stored per realization (e.g. in \
    `share/results/tables/compdat.csv`). This file can be exported to disk per realization by using
    the `ECL2CSV` forward model in ERT with subcommand `compdat`. [Link to ecl2csv compdat documentation.](https://equinor.github.io/ecl2df/usage/compdat.html)

    The connection status history of each cell is not necessarily complete in the `ecl2df` export,
    because status changes resulting from ACTIONs can't be extracted from the Eclipse input
    files. If the `ecl2df` export is good, it is recommended to use that. This will often be the
    case for history runs. But if not, an alternative way of extracting the data is described in
    the next section.

    **Well Connection status input**

    The `well_connection_status_file` is a path to a file stored per realization (e.g. in \
    `share/results/tables/wellconnstatus.csv`. This file can be exported to disk per realization
    by using the `ECL2CSV` forward model in ERT with subcommand `wellconnstatus`.  [Link to ecl2csv wellconnstatus documentation.](https://equinor.github.io/ecl2df/usage/wellconnstatus.html)

    This approach uses the CPI summary data to create a well connection status history: for
    each well connection cell there is one line for each time the connection is opened or closed.
    This data is sparse, but be aware that the CPI summary data can potentially become very large.

    **Zone layer mapping**

    The `zone_layer_mapping_file` file can be dumped to disk per realization by an internal \
    RMS script as part of the FMU workflow. A sample python script should be available in the \
    Drogon project.

    The file needs to be on the lyr format used by ResInsight:
    [Link to description of lyr format](https://resinsight.org/3d-main-window/formations/#formation-names-description-files-_lyr_).

    Zone colors can be specified in the lyr file, but only 6 digit hexadecimal codes will be used.

    If no file exists, layers will be used as zones.

    **Stratigraphy file**

    The `stratigraphy_file` file is intended to be generated per realization by an internal \
    RMS script as part of the FMU workflow, but can also be set up manually and copied to each
    realization. The stratigraphy is a tree structure, where each node has a name, an optional
    `color` parameter, and an optional `subzones` parameter which itself is a list of the same format.
    ```json
    [
        {
            "name": "ZoneA",
            "color": "#FFFFFF",
            "subzones": [
                {
                    "name": "ZoneA.1"
                },
                {
                    "name": "ZoneA.2"
                }
            ]
        },
        {
            "name": "ZoneB",
            "color": "#FFF000",
            "subzones": [
                {
                    "name": "ZoneB.1",
                    "color": "#FFF111"
                },
                {
                    "name": "ZoneB.2",
                    "subzones: {"name": "ZoneB.2.2"}
                }
            ]
        },
    ]
    ```
    The `stratigraphy_file` and the `zone_layer_mapping_file` will be combined to create the final \
    stratigraphy. A node will be removed if the name or any of the subnode names are not \
    present in the zone layer mapping. A Value Error is raised if any zones are present in the
    zone layer mapping but not in the stratigraphy.

    Colors can be supplied both trough the stratigraphy and through the zone_layer_mapping. \
    The following prioritization will be applied:
    1. Colors specified in the stratigraphy
    2. Colors specified in the zone layer mapping lyr file
    3. If none of the above is specified, theme colors will be added to the leaves of the tree

    **Well Attributes file**

    The `well_attributes_file` file is intended to be generated per realization by an internal \
    RMS script as part of the FMU workflow. A sample script will be made available, but it is \
    possible to manually set up the file and copy it to the correct folder on the scratch disk.\
    The categorical well attributes are completely flexible.

    The file should be a `.json` file on the following format:
    ```json
    {
        "version" : "0.1",
        "wells" : [
            {
                "alias" : {
                    "eclipse" : "OP_1"
                },
                "attributes" : {
                    "mlt_singlebranch" : "mlt",
                    "structure" : "East",
                    "welltype" : "producer"
                },
                "name" : "OP_1"
            },
            {
                "alias" : {
                    "eclipse" : "GI_1"
                },
                "attributes" : {
                    "mlt_singlebranch" : "singlebranch",
                    "structure" : "West",
                    "welltype" : "gas injector"
                },
                "name" : "GI_1"
            },
        ]
    }
    ```

    **KH unit**

    If defaulted, the plugin will look for the unit system of the Eclipse deck in the DATA file. \
    The kh unit will be deduced from the unit system, e.g. mD·m if METRIC.

    """

    def __init__(
        # pylint: disable=too-many-arguments
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        ensembles: list,
        compdat_file: str = "share/results/tables/compdat.csv",
        well_connection_status_file: str = "share/results/tables/wellconnstatus.csv",
        zone_layer_mapping_file: str = "rms/output/zone/simgrid_zone_layer_mapping.lyr",
        stratigraphy_file: str = "rms/output/zone/stratigraphy.json",
        well_attributes_file: str = "rms/output/wells/well_attributes.json",
        kh_unit: str = None,
        kh_decimal_places: int = 2,
    ):
        super().__init__()

        self._data_models = {
            ensemble: WellCompletionsDataModel(
                ensemble_name=ensemble,
                ensemble_path=webviz_settings.shared_settings["scratch_ensembles"][
                    ensemble
                ],
                compdat_file=compdat_file,
                well_connection_status_file=well_connection_status_file,
                zone_layer_mapping_file=zone_layer_mapping_file,
                stratigraphy_file=stratigraphy_file,
                well_attributes_file=well_attributes_file,
                kh_unit=kh_unit,
                kh_decimal_places=kh_decimal_places,
                theme_colors=webviz_settings.theme.plotly_theme["layout"]["colorway"],
            )
            for ensemble in ensembles
        }

        self.set_callbacks(app)

    def add_webvizstore(self) -> List[Tuple[Callable, List[Dict]]]:
        return [data_model.webviz_store for _, data_model in self._data_models.items()]

    @property
    def tour_steps(self) -> list:
        return layout_tour_steps(self.uuid)

    @property
    def layout(self) -> html.Div:
        return main_layout(
            get_uuid=self.uuid,
            ensembles=list(self._data_models.keys()),
        )

    def set_callbacks(self, app: Dash) -> None:
        plugin_callbacks(app, self.uuid, self._data_models)



from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import pandas as pd
from dash import dcc
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.deprecation_decorators import deprecated_plugin_arguments
from webviz_config.webviz_instance_info import WEBVIZ_INSTANCE_INFO, WebvizRunMode

from webviz_subsurface._models import (
    EnsembleSetModel,
    caching_ensemble_set_model_factory,
)
from webviz_subsurface._providers import (
    EnsembleSummaryProviderFactory,
    EnsembleTableProvider,
    EnsembleTableProviderFactory,
    Frequency,
)

from .controllers import parameter_qc_controller, parameter_response_controller
from .data_loaders import read_csv
from .models import (
    ParametersModel,
    ProviderTimeSeriesDataModel,
    SimulationTimeSeriesModel,
)
from .views import main_view


def check_deprecation_argument(
    csvfile_parameters: Optional[Path], csvfile_smry: Optional[Path]
) -> Optional[Tuple[str, str]]:
    if any(elm is not None for elm in [csvfile_parameters, csvfile_smry]):
        return (
            "The usage of aggregated csvfiles as user input options are deprecated. "
            "Please provide feedback if you see a need for a continuation "
            "of this functionality ",
            "",
        )
    return None


class ParameterAnalysis(WebvizPluginABC):
    """This plugin visualizes parameter distributions and statistics.
for FMU ensembles, and can be used to investigate parameter correlations
on reservoir simulation time series data.

---

**Using raw ensemble data stored in realization folders**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`rel_file_pattern`:** path to `.arrow` files with summary data.
* **`time_index`:** Time separation between extracted values. Can be e.g. `monthly` (default) or \
    `yearly`.
* **`drop_constants`:** Bool used to determine if constant parameters should be dropped. \
    Default is True.
* **`column_keys`:** List of vectors to extract. If not given, all vectors \
    from the simulations will be extracted. Wild card asterisk `*` can be used.
---

?> `Arrow` format for simulation time series data can be generated using the `ECL2CSV` forward \
model in ERT. On existing ensembles the command line tool `smry2arrow_batch` can be used to \
generate arrow files.

!> For smry data it is **strongly recommended** to keep the data frequency to a regular frequency \
(like `monthly` or `yearly`). This is because the statistics are calculated \
per DATE over all realizations in an ensemble, and the available dates should therefore not \
differ between individual realizations of an ensemble.

?> Vectors that are identified as historical vectors (e.g. FOPTH is the history of FOPT) will \
be plotted together with their non-historical counterparts as reference lines.

!> Parameter values are extracted automatically from the `parameters.txt` files in the individual
realizations if you have defined `ensembles`.


"""

    # pylint: disable=too-many-arguments, too-many-locals
    @deprecated_plugin_arguments(check_deprecation_argument)
    def __init__(
        self,
        app,
        webviz_settings: WebvizSettings,
        ensembles: Optional[list] = None,
        time_index: str = "monthly",
        column_keys: Optional[list] = None,
        drop_constants: bool = True,
        rel_file_pattern: str = "share/results/unsmry/*.arrow",
        csvfile_parameters: Path = None,
        csvfile_smry: Path = None,
    ):
        super().__init__()

        self.theme = webviz_settings.theme
        self.ensembles = ensembles
        self.vmodel: Optional[
            Union[SimulationTimeSeriesModel, ProviderTimeSeriesDataModel]
        ] = None

        run_mode_portable = WEBVIZ_INSTANCE_INFO.run_mode == WebvizRunMode.PORTABLE
        table_provider_factory = EnsembleTableProviderFactory.instance()

        if ensembles is not None:
            ensemble_paths = {
                ensemble_name: webviz_settings.shared_settings["scratch_ensembles"][
                    ensemble_name
                ]
                for ensemble_name in ensembles
            }

            resampling_frequency = Frequency(time_index)
            provider_factory = EnsembleSummaryProviderFactory.instance()

            try:
                provider_set = {
                    ens: provider_factory.create_from_arrow_unsmry_presampled(
                        str(ens_path), rel_file_pattern, resampling_frequency
                    )
                    for ens, ens_path in ensemble_paths.items()
                }
                self.vmodel = ProviderTimeSeriesDataModel(
                    provider_set=provider_set, column_keys=column_keys
                )

                parameter_df = create_df_from_table_provider(
                    provider_set={
                        ens_name: table_provider_factory.create_from_per_realization_parameter_file(
                            ens_path
                        )
                        for ens_name, ens_path in ensemble_paths.items()
                    }
                )

            except ValueError as error:
                message = (
                    f"Some/all ensembles are missing arrow files at {rel_file_pattern}.\n"
                    "If no arrow files have been generated with `ERT` using `ECL2CSV`, "
                    "the commandline tool `smry2arrow_batch` can be used to generate arrow "
                    "files for an ensemble"
                )
                if not run_mode_portable:
                    raise ValueError(message) from error

                # NOTE: this part below is to ensure backwards compatibility for portable app's
                # created before the arrow support. It should be removed in the future.
                emodel: EnsembleSetModel = (
                    caching_ensemble_set_model_factory.get_or_create_model(
                        ensemble_paths=ensemble_paths,
                        time_index=time_index,
                        column_keys=column_keys,
                    )
                )
                self.vmodel = SimulationTimeSeriesModel(
                    dataframe=emodel.get_or_load_smry_cached()
                )
                parameter_df = emodel.load_parameters()

        elif csvfile_parameters is None:
            raise ValueError("Either ensembles or csvfile_parameters must be specified")
        else:
            # NOTE: the try/except is for backwards compatibility with existing portable app's.
            # It should be removed in the future together with the support of aggregated csv-files
            try:
                parameter_df = create_df_from_table_provider(
                    provider_set=(
                        table_provider_factory.create_provider_set_from_aggregated_csv_file(
                            csvfile_parameters
                        )
                    )
                )
            except FileNotFoundError:
                if not run_mode_portable:
                    raise
                parameter_df = read_csv(csvfile_parameters)

            if csvfile_smry is not None:
                try:
                    smry_df = create_df_from_table_provider(
                        provider_set=(
                            table_provider_factory.create_provider_set_from_aggregated_csv_file(
                                csvfile_smry
                            )
                        )
                    )
                except FileNotFoundError:
                    if not run_mode_portable:
                        raise
                    smry_df = read_csv(csvfile_smry)

                self.vmodel = SimulationTimeSeriesModel(dataframe=smry_df)

        self.pmodel = ParametersModel(
            dataframe=parameter_df,
            theme=self.theme,
            drop_constants=drop_constants,
        )

        self.set_callbacks(app)

    @property
    def layout(self) -> dcc.Tabs:
        return main_view(
            get_uuid=self.uuid,
            vectormodel=self.vmodel,
            parametermodel=self.pmodel,
            theme=self.theme,
        )

    def set_callbacks(self, app) -> None:
        parameter_qc_controller(app=app, get_uuid=self.uuid, parametermodel=self.pmodel)
        if self.vmodel is not None:
            parameter_response_controller(
                app=app,
                get_uuid=self.uuid,
                parametermodel=self.pmodel,
                vectormodel=self.vmodel,
            )


def create_df_from_table_provider(
    provider_set: Dict[str, EnsembleTableProvider]
) -> pd.DataFrame:
    """Aggregates parameters from all ensemble into a common dataframe."""
    dfs = []
    for ens_name, provider in provider_set.items():
        df = provider.get_column_data(column_names=provider.column_names())
        df["ENSEMBLE"] = ens_name
        dfs.append(df)
    return pd.concat(dfs)



from pathlib import Path
from typing import Dict

from webviz_subsurface._providers import (
    EnsembleSummaryProvider,
    EnsembleSummaryProviderFactory,
    Frequency,
)

from .ensemble_summary_provider_set import EnsembleSummaryProviderSet


def create_lazy_ensemble_summary_provider_set_from_paths(
    name_path_dict: Dict[str, Path],
    rel_file_pattern: str,
) -> EnsembleSummaryProviderSet:
    """Create set of ensemble summary providers with lazy (on-demand) resampling/interpolation,
    from dictionary of ensemble name and corresponding arrow file paths

    `Input:`
    * name_path_dict: Dict[str, Path] - ensemble name as key and arrow file path as value
    * rel_file_pattern: str - specify a relative (per realization) file pattern to find the
    wanted .arrow files within each realization

    `Return:`
    Provider set with ensemble summary providers with lazy (on-demand) resampling/interpolation
    """
    provider_factory = EnsembleSummaryProviderFactory.instance()
    provider_dict: Dict[str, EnsembleSummaryProvider] = {}
    for name, path in name_path_dict.items():
        provider_dict[name] = provider_factory.create_from_arrow_unsmry_lazy(
            str(path), rel_file_pattern
        )
    return EnsembleSummaryProviderSet(provider_dict)


def create_presampled_ensemble_summary_provider_set_from_paths(
    name_path_dict: Dict[str, Path],
    rel_file_pattern: str,
    presampling_frequency: Frequency,
) -> EnsembleSummaryProviderSet:
    """Create set of ensemble summary providers without lazy resampling, but with specified
    frequency, from dictionary of ensemble name and corresponding arrow file paths

    `Input:`
    * name_path_dict: Dict[str, Path] - ensemble name as key and arrow file path as value
    * rel_file_pattern: str - specify a relative (per realization) file pattern to find the
    wanted .arrow files within each realization
    * presampling_frequency: Frequency - Frequency to sample input data in factory with, during
    import.

    `Return:`
    Provider set with ensemble summary providers with presampled data according to specified
    presampling frequency.
    """
    # TODO: Make presampling_frequency: Optional[Frequency] when allowing raw data for plugin
    provider_factory = EnsembleSummaryProviderFactory.instance()
    provider_dict: Dict[str, EnsembleSummaryProvider] = {}
    for name, path in name_path_dict.items():
        provider_dict[name] = provider_factory.create_from_arrow_unsmry_presampled(
            str(path), rel_file_pattern, presampling_frequency
        )
    return EnsembleSummaryProviderSet(provider_dict)



from typing import List, Optional

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from webviz_subsurface._figures import create_figure
from webviz_subsurface._utils.colors import hex_to_rgb, rgb_to_str, scale_rgb_lightness

from ._business_logic import QcFlags


def axis_defaults(showgrid: bool = True) -> dict:
    return {
        "showline": True,
        "linewidth": 2,
        "linecolor": "black",
        "mirror": True,
        "showgrid": showgrid,
        "gridwidth": 1,
        "gridcolor": "lightgrey",
    }


class WaterfallPlot:
    # Ensure fixed order of plot elements:
    ORDER = [
        "SWATINIT_WVOL",
        QcFlags.SWL_TRUNC.value,
        QcFlags.PPCWMAX.value,
        QcFlags.FINE_EQUIL.value,
        QcFlags.HC_BELOW_FWL.value,
        QcFlags.SWATINIT_1.value,
        "SWAT_WVOL",
    ]
    MEASURES = [
        "absolute",
        "relative",
        "relative",
        "relative",
        "relative",
        "relative",
        "total",
    ]

    def __init__(self, qc_vols: dict) -> None:
        # collect necessary values from input and make volume values more human friendly
        self.qc_vols = {
            key: (qc_vols[key] / (10**6))
            for key in self.ORDER + ["SWATINIT_HCVOL", "SWAT_HCVOL"]
        }
        self.qc_vols.update(
            {key: qc_vols[key] for key in ["WVOL_DIFF_PERCENT", "HCVOL_DIFF_PERCENT"]}
        )

    @property
    def range(self) -> list:
        range_min = min(self.qc_vols["SWATINIT_WVOL"], self.qc_vols["SWAT_WVOL"]) * 0.95
        range_max = max(self.qc_vols["SWATINIT_WVOL"], self.qc_vols["SWAT_WVOL"]) * 1.05
        return [range_min, range_max]

    @property
    def figure(self) -> go.Figure:
        return (
            go.Figure(
                go.Waterfall(
                    orientation="v",
                    measure=self.MEASURES,
                    x=self.ORDER,
                    textposition="outside",
                    text=self.create_bartext(),
                    y=[self.qc_vols[key] for key in self.ORDER],
                    connector={"mode": "spanning"},
                )
            )
            .update_yaxes(
                title="Water Volume (Mrm3)", range=self.range, **axis_defaults()
            )
            .update_xaxes(
                type="category",
                tickangle=-45,
                tickfont_size=17,
                **axis_defaults(showgrid=False),
            )
            .update_layout(
                plot_bgcolor="white",
                title="Waterfall chart showing changes from SWATINIT to SWAT",
                margin={"t": 50, "b": 50, "l": 50, "r": 50},
            )
        )

    def create_bartext(self) -> List[str]:
        """
        Create bartext for each qc_flag category with Water and HC volume change
        relative to SWATINIT_WVOL in percent.
        """
        text = []
        for bar_name in self.ORDER:
            bartext = [f"{self.qc_vols[bar_name]:.2f} Mrm3"]
            if bar_name != self.ORDER[0]:
                bartext.append(
                    f"Water {self.get_water_diff_in_percent(bar_name):.1f} %"
                )
                bartext.append(f"HC {self.get_hc_diff_in_percent(bar_name):.1f} %")

            text.append("<br>".join(bartext))
        return text

    def get_water_diff_in_percent(self, bar_name: str) -> float:
        if bar_name == self.ORDER[-1]:
            return self.qc_vols["WVOL_DIFF_PERCENT"]
        return (self.qc_vols[bar_name] / self.qc_vols["SWATINIT_WVOL"]) * 100

    def get_hc_diff_in_percent(self, bar_name: str) -> float:
        if bar_name == self.ORDER[-1]:
            return self.qc_vols["HCVOL_DIFF_PERCENT"]
        if self.qc_vols["SWATINIT_HCVOL"] > 0:
            return (-self.qc_vols[bar_name] / self.qc_vols["SWATINIT_HCVOL"]) * 100
        return 0


class PropertiesVsDepthSubplots:
    def __init__(
        self,
        dframe: pd.DataFrame,
        color_by: str,
        colormap: dict,
        discrete_color: bool = True,
    ) -> None:
        self.dframe = dframe
        self.color_by = color_by
        self.discrete_color = discrete_color
        self.colormap = colormap
        self.layout = [(1, 1), (1, 2), (2, 1), (2, 2)]
        self.responses = ["SWATINIT", "SWAT", "PRESSURE", "PC"]
        self.hover_data = self.create_hover_data(
            include_columns=["QC_FLAG", "EQLNUM", "SATNUM", "I", "J", "K"]
        )
        self.uirevision = "".join([str(x) for x in self.dframe["EQLNUM"].unique()])

        self._figure = self.create_empty_subplots_figure(rows=2, cols=2)
        self.add_subplotfigures_to_main_figure()
        self.add_contacts_to_plot()

    @property
    def figure(self) -> go.Figure:
        return self._figure

    @property
    def hovertemplate(self) -> go.Figure:
        return (
            "X: %{x}<br>Y: %{y}<br>"
            + "<br>".join(
                [
                    f"{col}:%{{customdata[{idx}]}}"
                    for idx, col in enumerate(self.hover_data)
                ]
            )
            + "<extra></extra>"
        )

    def create_empty_subplots_figure(self, rows: int, cols: int) -> go.Figure:
        return (
            make_subplots(
                rows=rows,
                cols=cols,
                subplot_titles=[f"Depth vs {resp}" for resp in self.responses],
                shared_yaxes=True,
                vertical_spacing=0.07,
                horizontal_spacing=0.05,
            )
            .update_layout(
                plot_bgcolor="white",
                uirevision=self.uirevision,
                margin={"t": 50, "b": 10, "l": 10, "r": 10},
                legend={"orientation": "h"},
                clickmode="event+select",
                coloraxis={"colorscale": "Viridis", **self.colorbar},
            )
            .update_yaxes(autorange="reversed", **axis_defaults())
            .update_xaxes(axis_defaults())
            .update_xaxes(row=1, col=2, matches="x")
        )

    def add_subplotfigures_to_main_figure(self) -> None:
        # for discrete colors there should be one trace per unique color
        unique_traces = (
            self.dframe[self.color_by].unique()
            if self.discrete_color
            else [self.color_by]
        )

        for color in unique_traces:
            df = self.dframe
            df = df[df[self.color_by] == color] if self.discrete_color else df
            customdata = np.stack([df[col] for col in self.hover_data], axis=-1)

            for idx, response in enumerate(self.responses):
                trace = go.Scattergl(
                    x=df[response],
                    y=df["Z"],
                    mode="markers",
                    name=color,
                    showlegend=idx == 0,
                    marker=self.set_marker_style(color, df),
                    unselected={"marker": self.set_unselected_marker_style(color)},
                    customdata=customdata,
                    hovertemplate=self.hovertemplate,
                    legendgroup=color,
                ).update(marker_size=10)

                row, col = self.layout[idx]
                self._figure.add_trace(trace, row=row, col=col)

    @property
    def colorbar(self) -> dict:
        if self.color_by != "PERMX":
            return {}
        tickvals = list(range(-4, 5, 1))
        return {
            "colorbar": {
                "tickvals": tickvals,
                "ticktext": [10**val for val in tickvals],
            }
        }

    def create_hover_data(self, include_columns: list) -> list:
        # ensure the colorby is the first entry in the list -> used in customdata in callback
        hover_data = [self.color_by]
        for col in include_columns:
            if col not in hover_data:
                hover_data.append(col)
        return hover_data

    def set_marker_style(self, color: str, df: pd.DataFrame) -> dict:
        if not self.discrete_color:
            return {
                "coloraxis": "coloraxis",
                "color": df[self.color_by]
                if self.color_by != "PERMX"
                else np.log10(df[self.color_by]),
            }
        return {"color": self.colormap[color], "opacity": 0.5}

    def set_unselected_marker_style(self, color: str) -> dict:
        if not self.discrete_color:
            return {"opacity": 0.1}
        return {
            "color": rgb_to_str(
                scale_rgb_lightness(hex_to_rgb(self.colormap[color]), 250)
            )
        }

    def add_contacts_to_plot(self) -> None:
        """Annotate axes with named horizontal lines for contacts."""
        for contact in ["OWC", "GWC", "GOC"]:
            if contact in self.dframe and self.dframe["EQLNUM"].nunique() == 1:
                # contacts are assumed constant in the dataframe
                value = self.dframe[contact].values[0]
                # do not include dummy contacts (shallower than the dataset)
                if value > self.dframe["Z"].min():
                    self._figure.add_hline(
                        value,
                        line={"color": "black", "dash": "dash", "width": 1.5},
                        annotation_text=f"{contact}={value:g}",
                        annotation_position="bottom left",
                    )


class MapFigure:
    def __init__(
        self,
        dframe: pd.DataFrame,
        color_by: str,
        colormap: dict,
        faultlinedf: Optional[pd.DataFrame] = None,
    ):
        self.dframe = dframe
        self.color_by = color_by
        self.colormap = colormap
        self.hover_data = ["I", "J", "K"]

        self._figure = self.create_figure()

        if faultlinedf is not None:
            self.add_fault_lines(faultlinedf)

    @property
    def figure(self) -> go.Figure:
        return self._figure

    @property
    def axis_layout(self) -> dict:
        return {
            "title": None,
            "showticklabels": False,
            "showgrid": False,
            "showline": False,
        }

    def create_figure(self) -> go.Figure:
        return (
            create_figure(
                plot_type="scatter",
                data_frame=self.dframe,
                x="X",
                y="Y",
                color=self.color_by
                if self.color_by != "PERMX"
                else np.log10(self.dframe[self.color_by]),
                color_discrete_map=self.colormap,
                xaxis={"constrain": "domain", **self.axis_layout},
                yaxis={"scaleanchor": "x", **self.axis_layout},
                hover_data=[self.color_by] + self.hover_data,
                color_continuous_scale="Viridis",
            )
            .update_traces(marker_size=10, unselected={"marker": {"opacity": 0}})
            .update_coloraxes(showscale=False)
            .update_layout(
                plot_bgcolor="white",
                margin={"t": 10, "b": 10, "l": 0, "r": 0},
                showlegend=False,
            )
        )

    def add_fault_lines(self, faultlinedf: pd.DataFrame) -> None:
        for _fault, faultdf in faultlinedf.groupby("POLY_ID"):
            self._figure.add_trace(
                {
                    "x": faultdf["X_UTME"],
                    "y": faultdf["Y_UTMN"],
                    "mode": "lines",
                    "type": "scatter",
                    "hoverinfo": "none",
                    "showlegend": False,
                    "line": {"color": "grey", "width": 1},
                }
            )



from typing import Dict, List, Tuple, Union
from uuid import uuid4

import numpy as np
import pandas as pd
import webviz_core_components as wcc
from dash import (
    ALL,
    Input,
    Output,
    State,
    callback,
    callback_context,
    dcc,
    html,
    no_update,
)
from dash.exceptions import PreventUpdate

from webviz_subsurface._models.parameter_model import ParametersModel
from webviz_subsurface._utils.formatting import printable_int_list


class ParameterFilter:
    """Component that can be added to a plugin to filter parameters"""

    def __init__(
        self,
        uuid: str,
        dframe: pd.DataFrame,
        reset_on_ensemble_update: bool = False,
        display_header: bool = True,
    ) -> None:
        """
        * **`uuid`:** Unique id (use the plugin id).
        * **`dframe`:** Dataframe, of all parameter values in all ensembles"""

        self._uuid = uuid
        self._reset_on_ensemble_update = reset_on_ensemble_update
        self._display_header = display_header
        self._pmodel = ParametersModel(
            dataframe=dframe,
            drop_constants=True,
            keep_numeric_only=False,
            drop_parameters_with_nan=True,
        )
        self._dframe = self._pmodel.dataframe
        self._range_parameters = self._get_range_parameters()
        self._column_precisions = self._get_column_precisions()
        self._dframe = self._dframe.round(self._column_precisions)
        self._discrete_parameters = self._get_discrete_parameters()
        self._min_max_ranges_per_ensemble = self._get_min_max_ranges_per_ensemble()
        self._min_max_all = self._return_min_max_ranges(self._ensembles)
        self.set_callbacks()

    @property
    def _ensembles(self) -> List[str]:
        return self._pmodel.ensembles

    @property
    def _realizations_per_ensemble(self) -> Dict[str, list]:
        return {
            ens: list(self._dframe[self._dframe["ENSEMBLE"] == ens]["REAL"].unique())
            for ens in self._ensembles
        }

    @property
    def is_sensitivity_run(self) -> bool:
        return self._pmodel.sensrun

    @property
    def _settings_visible(self) -> bool:
        return len(self._ensembles) > 1

    @property
    def _viewheight_parameters(self) -> str:
        return "60vh" if self._settings_visible else "75vh"

    @property
    def _parameters_per_ensemble(self) -> Dict[str, List]:
        return self._pmodel.parameters_per_ensemble

    @staticmethod
    def _precision(number: float) -> int:
        if abs(number) <= 0.01:
            return 6
        if abs(number) <= 1:
            return 3
        return 1

    def _get_range_parameters(self) -> List[str]:
        numeric_df = self._dframe.select_dtypes(include=np.number)
        return [
            col
            for col in numeric_df.columns[numeric_df.nunique() >= 6]
            if col not in self._pmodel.selectors
        ]

    def _get_discrete_parameters(self) -> List[str]:
        return [
            col
            for col in self._dframe.columns
            if col not in self._pmodel.selectors + self._range_parameters
        ]

    def _get_column_precisions(self) -> Dict[str, int]:
        return {
            col: self._precision(np.nanmin(self._dframe[col]))
            for col in self._range_parameters
        }

    def _get_min_max_ranges_per_ensemble(self) -> Dict[str, Dict[str, list]]:
        min_max_ranges = {}
        for ens in self._ensembles:
            dframe = self._dframe[self._dframe["ENSEMBLE"] == ens]
            parameters = self._parameters_per_ensemble[ens]
            min_max_ranges[ens] = {
                col: [dframe[col].min(), dframe[col].max()] for col in parameters
            }
        return min_max_ranges

    def _return_min_max_ranges(self, ensembles: list) -> Dict[str, list]:
        ranges = [self._min_max_ranges_per_ensemble[ens] for ens in ensembles]
        parameters = self._pmodel.get_parameters_for_ensembles(ensembles)
        return {
            col: [
                min(ens_ranges[col][0] for ens_ranges in ranges if col in ens_ranges),
                max(ens_ranges[col][1] for ens_ranges in ranges if col in ens_ranges),
            ]
            for col in parameters
        }

    def _missing_reals_in_ensemble(self, ensemble: str, reals: list) -> list:
        return [x for x in self._realizations_per_ensemble[ensemble] if x not in reals]

    @staticmethod
    def _get_values(
        sliders: list, selects: list, slider_id: list, select_id: list
    ) -> dict:
        values = {idv["name"]: val for val, idv in zip(sliders, slider_id)}
        values.update({idv["name"]: val for val, idv in zip(selects, select_id)})
        return values

    @property
    def layout(self) -> html.Div:
        children = [
            self.buttons,
            self.settings_layout,
            self.realizations_removed_layout,
            self.active_filters_layout,
            self.parameter_filter_layout,
        ]

        if self._display_header:
            children = [self.header] + children

        return html.Div(
            children=[
                html.Div(children=children),
                dcc.Store(
                    id={"id": self._uuid, "type": "data-store"},
                    data=self._initial_store,
                ),
                dcc.Store(id={"id": self._uuid, "type": "value-store"}, data={}),
                dcc.Store(id={"id": self._uuid, "type": "ensemble-update"}),
            ],
        )

    @property
    def _initial_store(self) -> Dict[str, List]:
        data = {}
        for ens_name, ens_df in self._dframe.groupby("ENSEMBLE"):
            data[ens_name] = sorted(list(ens_df["REAL"].unique()))
        return data

    @property
    def active_filters_layout(self) -> html.Details:
        return html.Details(
            className="webviz-selectors",
            children=[
                html.Summary(
                    "Active Filters",
                    className="webviz-underlined-label",
                    id={"id": self._uuid, "type": "filter-active-label"},
                ),
                html.Div(
                    style={"padding": "10px", "font-size": "15px"},
                    id={"id": self._uuid, "type": "filter-active-wrapper"},
                ),
            ],
        )

    @property
    def realizations_removed_layout(self) -> wcc.Selectors:
        return wcc.Selectors(
            label="Realizations removed",
            open_details=False,
            children=html.Div(id={"id": self._uuid, "type": "real-active-wrapper"}),
            style={"font-size": "15px"},
        )

    @property
    def parameter_filter_layout(self) -> wcc.Selectors:
        return wcc.Selectors(
            label="Parameters",
            children=html.Div(
                id={"id": self._uuid, "type": "filter-wrapper"},
                style={"overflowY": "auto", "height": self._viewheight_parameters},
            ),
        )

    @property
    def settings_layout(self) -> html.Div:
        return html.Div(
            wcc.Selectors(
                label="Settings",
                children=html.Div(
                    style={
                        "marginBottom": "5px",
                        "backgroundColor": "white",
                        "border": "1px solid #bbb",
                        "border-radius": "4px",
                        "padding": "5px",
                    },
                    children=[
                        wcc.Checklist(
                            id={"id": self._uuid, "type": "ensemble-selector"},
                            label="Filter Ensembles:",
                            options=[
                                {"label": val, "value": val} for val in self._ensembles
                            ],
                            value=self._ensembles,
                            vertical=False,
                        ),
                        wcc.RadioItems(
                            id={"id": self._uuid, "type": "range-selector"},
                            label="Use ranges from:",
                            options=[
                                {"label": "Selected", "value": "selected"},
                                {"label": "All", "value": "all"},
                            ],
                            value="selected",
                            vertical=False,
                        ),
                    ],
                ),
            ),
            style={"display": "block" if self._settings_visible else "none"},
        )

    @property
    def header(self) -> wcc.Header:
        return wcc.Header(
            "Parameter filter",
            style={"color": "black", "marginBottom": "15px", "borderColor": "black"},
        )

    @property
    def buttons(self) -> html.Div:
        button_style = {
            "height": "30px",
            "line-height": "30px",
            "background-color": "white",
        }
        return html.Div(
            style={"marginTop": "10px", "height": "35px"},
            children=[
                html.Button(
                    "Reset",
                    style={"width": "48%", "float": "right", **button_style},
                    id={"id": self._uuid, "type": "reset-button"},
                ),
                html.Button(
                    "Apply",
                    style={"width": "48%", "float": "left", **button_style},
                    id={"id": self._uuid, "type": "apply-button"},
                ),
            ],
        )

    # pylint: disable=too-many-statements
    def set_callbacks(self) -> None:
        @callback(
            Output({"id": self._uuid, "type": "data-store"}, "data"),
            Output({"id": self._uuid, "type": "value-store"}, "data"),
            Input({"id": self._uuid, "type": "apply-button"}, "n_clicks"),
            State({"id": self._uuid, "type": "ensemble-selector"}, "value"),
            State({"id": self._uuid, "type": "slider", "name": ALL, "r": ALL}, "value"),
            State({"id": self._uuid, "type": "select", "name": ALL, "r": ALL}, "value"),
            State({"id": self._uuid, "type": "data-store"}, "data"),
            State({"id": self._uuid, "type": "slider", "name": ALL, "r": ALL}, "id"),
            State({"id": self._uuid, "type": "select", "name": ALL, "r": ALL}, "id"),
        )
        def store_selections(
            n_clicks: int,
            ensembles: list,
            sliders: list,
            selects: list,
            data_store: dict,
            slider_id: List[dict],
            select_id: List[dict],
        ) -> tuple:
            if not n_clicks:
                raise PreventUpdate

            values = self._get_values(sliders, selects, slider_id, select_id)

            real_dict = {}
            for ens, ens_df in self._dframe.groupby("ENSEMBLE"):
                if ens in ensembles:
                    for col in self._parameters_per_ensemble[ens]:
                        if col in self._range_parameters:
                            ens_df = ens_df[
                                (ens_df[col] >= values[col][0])
                                & (ens_df[col] <= values[col][1])
                            ]
                        else:
                            ens_df = ens_df[ens_df[col].isin(values[col])]

                real_dict[ens] = list(ens_df["REAL"].unique())

            # only update data_store if realization filter has changed
            update_data_store = any(
                real_dict[ens] != data_store[ens] for ens in self._ensembles
            )
            return real_dict if update_data_store else no_update, values

        @callback(
            Output({"id": self._uuid, "type": "filter-wrapper"}, "children"),
            Output({"id": self._uuid, "type": "apply-button"}, "n_clicks"),
            Input({"id": self._uuid, "type": "reset-button"}, "n_clicks"),
            Input({"id": self._uuid, "type": "range-selector"}, "value"),
            Input({"id": self._uuid, "type": "ensemble-selector"}, "value"),
            State({"id": self._uuid, "type": "slider", "name": ALL, "r": ALL}, "value"),
            State({"id": self._uuid, "type": "slider", "name": ALL, "r": ALL}, "id"),
            State({"id": self._uuid, "type": "select", "name": ALL, "r": ALL}, "value"),
            State({"id": self._uuid, "type": "select", "name": ALL, "r": ALL}, "id"),
        )
        # pylint: disable=too-many-locals
        def update_filtercomponents_and_apply(
            _reset_click: int,
            range_from: str,
            ensembles: list,
            sliders: list,
            slider_id: List[dict],
            selects: list,
            select_id: List[dict],
        ) -> Tuple[list, int]:
            ctx = callback_context.triggered[0]["prop_id"]
            if not ensembles:
                raise PreventUpdate

            reset = (
                self._reset_on_ensemble_update and "ensemble-selector" in ctx
            ) or "reset" in ctx
            if not reset and (range_from == "all" and "ensemble-selector" in ctx):
                return no_update, 1

            values = self._get_values(sliders, selects, slider_id, select_id)
            min_max_range = self._return_min_max_ranges(ensembles)
            ens_df = self._dframe[self._dframe["ENSEMBLE"].isin(ensembles)]

            children = []
            for col in self._pmodel.get_parameters_for_ensembles(ensembles):
                if col in self._range_parameters:
                    children.append(
                        make_range_slider(
                            min_max=(
                                self._min_max_all[col]
                                if range_from == "all"
                                else min_max_range[col]
                            ),
                            value=min_max_range[col]
                            if reset or col not in values
                            else values[col],
                            name=col,
                            uuid=self._uuid,
                            step=10 ** -self._column_precisions[col],
                        )
                    )

                else:
                    df = self._dframe if range_from == "all" else ens_df
                    unique_values = list(df[col].dropna().unique())
                    value = (
                        list(ens_df[col].dropna().unique())
                        if reset or col not in values
                        else [x for x in values[col] if x in unique_values]
                    )
                    children.append(
                        make_discrete_selector(
                            value=value if value else unique_values,
                            options=unique_values,
                            name=col,
                            uuid=self._uuid,
                        )
                    )

            return children, 1 if not "range-selector" in ctx else no_update

        @callback(
            Output({"id": self._uuid, "type": "real-active-wrapper"}, "children"),
            Input({"id": self._uuid, "type": "value-store"}, "data"),
            State({"id": self._uuid, "type": "data-store"}, "data"),
        )
        def _update_missing_real_text(_value_store: list, data_store: dict) -> list:
            return [
                html.Div(
                    [
                        html.Span(f"{ens}:  ", style={"font-weight": "bold"}),
                        printable_int_list(self._missing_reals_in_ensemble(ens, reals)),
                    ]
                )
                for ens, reals in data_store.items()
            ]

        @callback(
            Output({"id": self._uuid, "type": "filter-active-label"}, "style"),
            Output({"id": self._uuid, "type": "filter-active-wrapper"}, "children"),
            Input({"id": self._uuid, "type": "value-store"}, "data"),
            State({"id": self._uuid, "type": "ensemble-selector"}, "value"),
            State({"id": self._uuid, "type": "slider", "name": ALL, "r": ALL}, "value"),
            State({"id": self._uuid, "type": "select", "name": ALL, "r": ALL}, "value"),
            State({"id": self._uuid, "type": "slider", "name": ALL, "r": ALL}, "id"),
            State({"id": self._uuid, "type": "select", "name": ALL, "r": ALL}, "id"),
        )
        def _update_active_filter_text(
            _value_store: list,
            ensembles: list,
            sliders: list,
            selects: list,
            slider_id: List[dict],
            select_id: List[dict],
        ) -> Tuple[dict, list]:

            min_max = self._return_min_max_ranges(ensembles)
            df = self._dframe[self._dframe["ENSEMBLE"].isin(ensembles)]
            values = self._get_values(sliders, selects, slider_id, select_id)

            params_active = []
            for col in self._pmodel.get_parameters_for_ensembles(ensembles):
                if col in self._range_parameters:
                    min_val, max_val = values[col]
                    if (min_val > min_max[col][0]) or (max_val < min_max[col][1]):
                        params_active.append(
                            html.Div(f"{col}:", style={"font-weight": "bold"})
                        )
                        params_active.append(html.Div(f"{min_val} - {max_val}"))
                else:
                    unique_values = list(df[col].dropna().unique())
                    if not all(x in values[col] for x in unique_values):
                        params_active.append(html.Div(f"{col}"))

            active_filter_style = (
                {"color": "lightgray", "border-color": "lightgray"}
                if not params_active
                else {}
            )
            return active_filter_style, params_active

        @callback(
            Output({"id": self._uuid, "type": "apply-button"}, "style"),
            Input({"id": self._uuid, "type": "apply-button"}, "n_clicks"),
            Input({"id": self._uuid, "type": "slider", "name": ALL, "r": ALL}, "value"),
            Input({"id": self._uuid, "type": "select", "name": ALL, "r": ALL}, "value"),
            State({"id": self._uuid, "type": "slider", "name": ALL, "r": ALL}, "id"),
            State({"id": self._uuid, "type": "select", "name": ALL, "r": ALL}, "id"),
            State({"id": self._uuid, "type": "value-store"}, "data"),
            State({"id": self._uuid, "type": "apply-button"}, "style"),
        )
        def _update_apply_button_style(
            _n_clicks: int,
            sliders: List[list],
            selects: List[list],
            slider_id: List[dict],
            select_id: List[dict],
            values_stored: Dict[str, list],
            style: dict,
        ) -> dict:

            ctx = callback_context.triggered[0]["prop_id"]
            values = self._get_values(sliders, selects, slider_id, select_id)
            modified_values = any(
                values[col] != values_stored.get(col, values[col]) for col in values
            )

            if modified_values and not "apply-button" in ctx:
                style["background-color"] = "#7393B3"
                style["color"] = "white"
            else:
                style["background-color"] = "white"
                style["color"] = "#555"
            return style

        @callback(
            Output({"id": self._uuid, "type": "ensemble-selector"}, "value"),
            Output({"id": self._uuid, "type": "ensemble-selector"}, "options"),
            Input({"id": self._uuid, "type": "ensemble-update"}, "data"),
            prevent_initial_call=True,
        )
        def _update_ensembles_from_outside(ensembles: list) -> tuple:
            """Update ensemble in parameter filter"""
            return (
                ensembles,
                [
                    {
                        "label": val,
                        "value": val,
                        "disabled": val not in ensembles or len(ensembles) == 1,
                    }
                    for val in self._ensembles
                ],
            )


def make_range_slider(
    value: list, min_max: list, step: float, name: str, uuid: str
) -> html.Div:
    return wcc.RangeSlider(
        label=name,
        id={
            "id": uuid,
            "type": "slider",
            "name": name,
            "r": str(uuid4()),  # needed to trigger update (bug)
        },
        min=min_max[0],
        max=min_max[1],
        step=step,
        value=_validate_slider_value(min_max, value),
        marks={format_marker_value(val): {"label": str(val)} for val in min_max},
        tooltip={"always_visible": False},
    )


def make_discrete_selector(
    value: list, options: list, name: str, uuid: str
) -> html.Div:
    return wcc.SelectWithLabel(
        label=name,
        id={
            "id": uuid,
            "type": "select",
            "name": name,
            "r": str(uuid4()),  # needed to trigger update (bug)
        },
        options=[{"label": val, "value": val} for val in sorted(options)],
        value=value,
        multi=True,
    )


def _validate_slider_value(min_max: list, value: list) -> list:
    min_val, max_val = min_max
    if min_val > value[0]:
        value[0] = min_val
    if max_val < value[0]:
        value[0] = max_val
    if max_val < value[1]:
        value[1] = max_val
    if min_val > value[1]:
        value[1] = min_val
    return value


def format_marker_value(number: Union[float, int]) -> str:
    # In order to show up on the RangeSlider the values used for the marks
    # needs to be without trailing zeros and scientific notification
    number = float(number)
    if number.is_integer():
        return str(int(number))
    return f"{number:.6f}".rstrip("0")



from typing import List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from .._types import SubplotGroupByOptions


class GroupBySettings(SettingsGroupABC):
    class Ids(StrEnum):
        SUBPLOT_OWNER_OPTIONS_RADIO_ITEMS = "subplot-owner-options-radio-items"

    def __init__(self) -> None:
        super().__init__("Group by")

    def layout(self) -> List[Component]:
        return [
            wcc.RadioItems(
                id=self.register_component_unique_id(
                    GroupBySettings.Ids.SUBPLOT_OWNER_OPTIONS_RADIO_ITEMS
                ),
                options=[
                    {
                        "label": "Time Series",
                        "value": SubplotGroupByOptions.VECTOR,
                    },
                    {
                        "label": "Ensemble",
                        "value": SubplotGroupByOptions.ENSEMBLE,
                    },
                ],
                value=SubplotGroupByOptions.VECTOR,
            )
        ]



from datetime import datetime
from typing import List, Optional, Sequence

import pandas as pd
import pytest
from pandas._testing import assert_frame_equal

from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.dataframe_utils import make_date_column_datetime_object

# pylint: disable=line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.history_vectors import (
    create_history_vectors_df,
)

from ....mocks.ensemble_summary_provider_dummy import EnsembleSummaryProviderDummy

# *******************************************************************
#####################################################################
#
# CONFIGURE TESTDATA
#
#####################################################################
# *******************************************************************


class EnsembleSummaryProviderMock(EnsembleSummaryProviderDummy):
    """Mock implementation of EnsembleSummaryProvider for testing of creating history
    vector dataframe

    Implements necessary methods for obtaining wanted test data
    """

    def __init__(self, input_df: pd.DataFrame) -> None:
        super().__init__()

        # Configure dataframe
        self._df = input_df
        self._vector_names: List[str] = list(
            set(self._df.columns) ^ set(["DATE", "REAL"])
        )
        self._realizations: List[int] = list(self._df["REAL"].unique())

    def supports_resampling(self) -> bool:
        return False

    def vector_names(self) -> List[str]:
        return self._vector_names

    def realizations(self) -> List[int]:
        return self._realizations

    def get_vectors_df(
        self,
        vector_names: Sequence[str],
        __resampling_frequency: Optional[Frequency],
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:
        if realizations:
            if not set(realizations).issubset(set(self._realizations)):
                raise ValueError(
                    "Requested realizations are not subset of provider realizations!"
                )
            return (
                self._df[["DATE", "REAL"] + list(vector_names)]
                .loc[self._df["REAL"] == realizations[0]]
                .reset_index(drop=True)
            )
        raise ValueError("Expected valid realizations argument for mock!")


# fmt: off
# Vector WA and WB, with corresponding historical vectors WAH and WBH, respectively
# NOTE:
#  - Method utilize guess of historical vector
#       - Assume vectors starting with "F", "G" or "W". Thereby "WA" and "WB" in test
#       - String split on ":" and append "H" on first part, i.e. WA -> WAH,
#         WA:OP_1 -> WAH:OP_1

INPUT_DF = pd.DataFrame(
    columns = ["DATE", "REAL",  "WA", "WAH", "WB", "WBH"],
    data = [
        [datetime(2000,1,1),  0,  11.0,  13.0,  15.0,   17.0],
        [datetime(2000,2,1),  0,  21.0,  23.0,  25.0,   27.0],
        [datetime(2000,3,1),  0,  31.0,  33.0,  35.0,   37.0],
        [datetime(2000,4,1),  0,  41.0,  43.0,  45.0,   47.0],
        [datetime(2000,5,1),  0,  51.0,  53.0,  55.0,   57.0],
        [datetime(2000,1,1),  3,  110.0, 115.0, 135.0,  139.0],
        [datetime(2000,2,1),  3,  310.0, 215.0, 235.0,  239.0],
        [datetime(2000,3,1),  3,  410.0, 315.0, 335.0,  339.0],
        [datetime(2000,4,1),  3,  510.0, 415.0, 435.0,  439.0],
        [datetime(2000,5,1),  3,  610.0, 515.0, 535.0,  539.0],
    ]
)
# Data of real = 0 for "WAH"
EXPECTED_WA_HISTORY_DF = pd.DataFrame(
    columns = ["DATE", "REAL", "WA"],
    data = [
        [datetime(2000,1,1),  0,  13.0],
        [datetime(2000,2,1),  0,  23.0],
        [datetime(2000,3,1),  0,  33.0],
        [datetime(2000,4,1),  0,  43.0],
        [datetime(2000,5,1),  0,  53.0],
    ]
)
# Data of real = 0 for "WBH" and "WAH"
EXPECTED_WB_WA_HISTORY_DF = pd.DataFrame(
    columns = ["DATE", "REAL", "WB", "WA"],
    data = [
        [datetime(2000,1,1),  0,  17.0,  13.0],
        [datetime(2000,2,1),  0,  27.0,  23.0],
        [datetime(2000,3,1),  0,  37.0,  33.0],
        [datetime(2000,4,1),  0,  47.0,  43.0],
        [datetime(2000,5,1),  0,  57.0,  53.0],
    ]
)
make_date_column_datetime_object(INPUT_DF)
make_date_column_datetime_object(EXPECTED_WA_HISTORY_DF)
make_date_column_datetime_object(EXPECTED_WB_WA_HISTORY_DF)


# Dates AFTER year 2262!
# NOTE: datetime.datetime after year 2262 is not converted to pd.Timestamp, thus
# no need to make date column datetime object
INPUT_YEAR_2265_DF = pd.DataFrame(
    columns = ["DATE", "REAL",  "WA", "WAH", "WB", "WBH"],
    data = [
        [datetime(2265,1,1),  4,  11.0,  13.0,  15.0,   17.0],
        [datetime(2265,2,1),  4,  21.0,  23.0,  25.0,   27.0],
        [datetime(2265,3,1),  4,  31.0,  33.0,  35.0,   37.0],
        [datetime(2265,4,1),  4,  41.0,  43.0,  45.0,   47.0],
        [datetime(2265,5,1),  4,  51.0,  53.0,  55.0,   57.0],
        [datetime(2265,1,1),  2,  110.0, 115.0, 135.0,  139.0],
        [datetime(2265,2,1),  2,  310.0, 215.0, 235.0,  239.0],
        [datetime(2265,3,1),  2,  410.0, 315.0, 335.0,  339.0],
        [datetime(2265,4,1),  2,  510.0, 415.0, 435.0,  439.0],
        [datetime(2265,5,1),  2,  610.0, 515.0, 535.0,  539.0],
    ]
)
# Data of real = 2 for "WAH"
EXPECTED_YEAR_2265_WA_HISTORY_DF = pd.DataFrame(
    columns = ["DATE", "REAL", "WA"],
    data = [
        [datetime(2265,1,1),  2,  115.0],
        [datetime(2265,2,1),  2,  215.0],
        [datetime(2265,3,1),  2,  315.0],
        [datetime(2265,4,1),  2,  415.0],
        [datetime(2265,5,1),  2,  515.0],
    ]
)
# Data of real = 2 for "WBH" and "WAH"
EXPECTED_YEAR_2265_WB_WA_HISTORY_DF = pd.DataFrame(
    columns = ["DATE", "REAL", "WB", "WA"],
    data = [
        [datetime(2265,1,1),  2,  139.0,  115.0],
        [datetime(2265,2,1),  2,  239.0,  215.0],
        [datetime(2265,3,1),  2,  339.0,  315.0],
        [datetime(2265,4,1),  2,  439.0,  415.0],
        [datetime(2265,5,1),  2,  539.0,  515.0],
    ]
)

# fmt: on


# *******************************************************************
#####################################################################
#
# UNIT TESTS
#
#####################################################################
# *******************************************************************


TEST_CASES = [
    pytest.param(
        EnsembleSummaryProviderMock(INPUT_DF),
        EXPECTED_WA_HISTORY_DF,
        EXPECTED_WB_WA_HISTORY_DF,
    ),
    pytest.param(
        EnsembleSummaryProviderMock(INPUT_YEAR_2265_DF),
        EXPECTED_YEAR_2265_WA_HISTORY_DF,
        EXPECTED_YEAR_2265_WB_WA_HISTORY_DF,
    ),
]


@pytest.mark.parametrize(
    "provider, expected_wa_history_df, expected_wb_wa_history_df", TEST_CASES
)
def test_create_history_vectors_df(
    provider: pd.DataFrame,
    expected_wa_history_df: pd.DataFrame,
    expected_wb_wa_history_df: pd.DataFrame,
) -> None:
    create_wa_history_df = create_history_vectors_df(provider, ["WA"], None)
    create_wb_wa_history_df = create_history_vectors_df(provider, ["WB", "WA"], None)

    assert_frame_equal(create_wa_history_df, expected_wa_history_df)
    assert_frame_equal(create_wb_wa_history_df, expected_wb_wa_history_df)



import webviz_core_components as wcc
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import ViewElementABC


class SubplotGraph(ViewElementABC):
    """View element for subplot graph"""

    class Ids(StrEnum):
        GRAPH = "graph"

    def __init__(self, height: str = "86vh") -> None:
        super().__init__()
        self.height = height

    def inner_layout(self) -> wcc.Graph:
        return wcc.Graph(
            style={"display": "block", "height": self.height},
            id=self.register_component_unique_id(SubplotGraph.Ids.GRAPH),
        )



import hashlib
import logging
import os
from pathlib import Path
from typing import Optional

from webviz_config.webviz_factory import WebvizFactory
from webviz_config.webviz_factory_registry import WEBVIZ_FACTORY_REGISTRY
from webviz_config.webviz_instance_info import WebvizRunMode

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._provider_impl_file import ProviderImplFile
from .well_provider import WellProvider

LOGGER = logging.getLogger(__name__)


class WellProviderFactory(WebvizFactory):
    def __init__(self, root_storage_folder: Path, allow_storage_writes: bool) -> None:
        self._storage_dir = Path(root_storage_folder) / __name__
        self._allow_storage_writes = allow_storage_writes

        LOGGER.info(f"WellProviderFactory init: storage_dir={self._storage_dir}")

        if self._allow_storage_writes:
            os.makedirs(self._storage_dir, exist_ok=True)

    @staticmethod
    def instance() -> "WellProviderFactory":
        """Static method to access the singleton instance of the factory."""

        factory = WEBVIZ_FACTORY_REGISTRY.get_factory(WellProviderFactory)
        if not factory:
            app_instance_info = WEBVIZ_FACTORY_REGISTRY.app_instance_info
            storage_folder = app_instance_info.storage_folder
            allow_writes = app_instance_info.run_mode != WebvizRunMode.PORTABLE

            factory = WellProviderFactory(storage_folder, allow_writes)

            # Store the factory object in the global factory registry
            WEBVIZ_FACTORY_REGISTRY.set_factory(WellProviderFactory, factory)

        return factory

    def create_from_well_files(
        self, well_folder: str, well_suffix: str, md_logname: Optional[str]
    ) -> WellProvider:
        timer = PerfTimer()

        file_pattern = str(Path(well_folder) / f"*{well_suffix}")
        storage_key = f"from_files__{_make_hash_string(f'{file_pattern}_{md_logname}')}"

        provider = ProviderImplFile.from_backing_store(self._storage_dir, storage_key)
        if provider:
            LOGGER.info(
                f"Loaded well provider from backing store in {timer.elapsed_s():.2f}s ("
                f"file_pattern={file_pattern})"
            )
            return provider

        # We can only import data from data source if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(f"Failed to load well provider for {file_pattern}")

        LOGGER.info(f"Importing/writing well data for: {file_pattern}")

        timer.lap_s()
        src_file_names = sorted(
            [str(filename) for filename in Path(well_folder).glob(f"*{well_suffix}")]
        )
        et_discover_s = timer.lap_s()

        ProviderImplFile.write_backing_store(
            self._storage_dir,
            storage_key,
            well_file_names=src_file_names,
            md_logname=md_logname,
        )
        et_write_s = timer.lap_s()

        provider = ProviderImplFile.from_backing_store(self._storage_dir, storage_key)
        if not provider:
            raise ValueError(f"Failed to load/create well provider for {file_pattern}")

        LOGGER.info(
            f"Saved well provider to backing store in {timer.elapsed_s():.2f}s ("
            f"discover={et_discover_s:.2f}s, write={et_write_s:.2f}s, file_pattern={file_pattern})"
        )

        return provider


def _make_hash_string(string_to_hash: str) -> str:
    # There is no security risk here and chances of collision should be very slim
    return hashlib.md5(string_to_hash.encode()).hexdigest()  # nosec



import warnings
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import xtgeo
from dash import Dash, Input, Output, State, callback_context, no_update
from dash._callback import NoUpdate
from dash.exceptions import PreventUpdate

from webviz_subsurface._datainput.well import (
    create_leaflet_well_marker_layer,
    make_well_layer,
)
from webviz_subsurface._models import SurfaceLeafletModel, SurfaceSetModel, WellSetModel


# pylint: disable=too-many-statements
def update_maps(
    app: Dash,
    get_uuid: Callable,
    surface_set_models: Dict[str, SurfaceSetModel],
    well_set_model: WellSetModel,
) -> None:
    @app.callback(
        Output({"id": get_uuid("map"), "element": "label"}, "children"),
        Output(get_uuid("leaflet-map1"), "layers"),
        Output({"id": get_uuid("map2"), "element": "label"}, "children"),
        Output(get_uuid("leaflet-map2"), "layers"),
        Output({"id": get_uuid("map3"), "element": "label"}, "children"),
        Output(get_uuid("leaflet-map3"), "layers"),
        Input(
            {
                "id": get_uuid("map-settings"),
                "map_id": "map1",
                "element": "surfaceattribute",
            },
            "value",
        ),
        Input(
            {
                "id": get_uuid("map-settings"),
                "map_id": "map2",
                "element": "surfaceattribute",
            },
            "value",
        ),
        Input(
            {
                "id": get_uuid("map-settings"),
                "map_id": "map1",
                "element": "surfacename",
            },
            "value",
        ),
        Input(
            {
                "id": get_uuid("map-settings"),
                "map_id": "map2",
                "element": "surfacename",
            },
            "value",
        ),
        Input(
            {"id": get_uuid("map-settings"), "map_id": "map1", "element": "ensemble"},
            "value",
        ),
        Input(
            {"id": get_uuid("map-settings"), "map_id": "map2", "element": "ensemble"},
            "value",
        ),
        Input(
            {
                "id": get_uuid("map-settings"),
                "map_id": "map1",
                "element": "calculation",
            },
            "value",
        ),
        Input(
            {
                "id": get_uuid("map-settings"),
                "map_id": "map2",
                "element": "calculation",
            },
            "value",
        ),
        Input(get_uuid("leaflet-map1"), "switch"),
        Input(get_uuid("leaflet-map2"), "switch"),
        Input(get_uuid("leaflet-map3"), "switch"),
        Input(
            {"id": get_uuid("map-settings"), "map_id": "map1", "element": "options"},
            "value",
        ),
        Input(
            {"id": get_uuid("map-settings"), "map_id": "map2", "element": "options"},
            "value",
        ),
        Input(get_uuid("map-color-ranges"), "data"),
        Input(
            {"id": get_uuid("map-settings"), "settings": "compute_diff"},
            "value",
        ),
        Input(get_uuid("realization-store"), "data"),
        Input({"id": get_uuid("intersection-data"), "element": "well"}, "value"),
        Input({"id": get_uuid("map"), "element": "stored_polyline"}, "data"),
        Input({"id": get_uuid("map"), "element": "stored_xline"}, "data"),
        Input({"id": get_uuid("map"), "element": "stored_yline"}, "data"),
        Input({"id": get_uuid("intersection-data"), "element": "source"}, "value"),
        State(get_uuid("leaflet-map1"), "layers"),
        State(get_uuid("leaflet-map2"), "layers"),
        State(get_uuid("leaflet-map3"), "layers"),
    )
    # pylint: disable=too-many-arguments, too-many-locals, too-many-branches
    def _update_maps(
        surfattr_map: str,
        surfattr_map2: str,
        surfname_map: str,
        surfname_map2: str,
        ensemble_map: str,
        ensemble_map2: str,
        calc_map: str,
        calc_map2: str,
        shade_map: Dict[str, bool],
        shade_map2: Dict[str, bool],
        shade_map3: Dict[str, bool],
        options: List[str],
        options2: List[str],
        color_range_settings: Dict,
        compute_diff: List[str],
        real_list: List[str],
        wellname: Optional[str],
        polyline: Optional[List],
        xline: Optional[List],
        yline: Optional[List],
        source: str,
        current_map: List,
        current_map2: List,
        current_map3: List,
    ) -> Tuple[str, List, str, List, str, List]:
        """Generate Leaflet layers for the three map views"""
        realizations = [int(real) for real in real_list]
        ctx = callback_context.triggered[0]
        if "compute_diff" in ctx["prop_id"]:
            if not compute_diff:
                return (
                    no_update,
                    no_update,
                    no_update,
                    no_update,
                    no_update,
                    [],
                )

        # Check if map is already generated and should just be updated with polylines
        update_poly_only = bool(
            current_map
            and (
                "stored_polyline" in ctx["prop_id"]
                or "stored_yline" in ctx["prop_id"]
                or "stored_xline" in ctx["prop_id"]
            )
        )
        if polyline is not None:
            poly_layer = create_leaflet_polyline_layer(
                polyline, name="Polyline", poly_id="random_line"
            )
            for map_layers in [current_map, current_map2, current_map3]:
                map_layers = replace_or_add_map_layer(
                    map_layers, "Polyline", poly_layer
                )
        if xline is not None and source == "xline":
            xline_layer = create_leaflet_polyline_layer(
                xline, name="Xline", poly_id="x_line"
            )
            for map_layers in [current_map, current_map2, current_map3]:
                map_layers = replace_or_add_map_layer(map_layers, "Xline", xline_layer)
        if yline is not None and source == "yline":
            yline_layer = create_leaflet_polyline_layer(
                yline, name="Yline", poly_id="y_line"
            )
            for map_layers in [current_map, current_map2, current_map3]:
                map_layers = replace_or_add_map_layer(map_layers, "Yline", yline_layer)
        # If callback is triggered by polyline drawing, only update polyline
        if update_poly_only:
            return (
                f"Surface A: {surfattr_map} - {surfname_map} - {ensemble_map} - {calc_map}",
                current_map,
                f"Surface B: {surfattr_map2} - {surfname_map2} - {ensemble_map2} - {calc_map2}",
                no_update,
                "Surface A-B",
                no_update,
            )

        if wellname is not None:
            well = well_set_model.get_well(wellname)
            well_layer = make_well_layer(well, name=well.name)

            # If callback is triggered by well change, only update well layer
            if "well" in ctx["prop_id"] or (
                "source" in ctx["prop_id"] and source == "well"
            ):
                for map_layers in [current_map, current_map2, current_map3]:
                    map_layers = replace_or_add_map_layer(
                        map_layers, "Well", well_layer
                    )
                return (
                    f"Surface A: {surfattr_map} - {surfname_map} - "
                    f"{ensemble_map} - {calc_map}",
                    current_map,
                    f"Surface B: {surfattr_map2} - {surfname_map2} - "
                    f"{ensemble_map2} - {calc_map2}",
                    current_map2,
                    "Surface A-B",
                    no_update,
                )

        # Calculate maps
        if calc_map in ["Mean", "StdDev", "Max", "Min", "P90", "P10"]:
            surface = surface_set_models[ensemble_map].calculate_statistical_surface(
                name=surfname_map,
                attribute=surfattr_map,
                calculation=calc_map,
                realizations=realizations,
            )
        else:
            surface = surface_set_models[ensemble_map].get_realization_surface(
                name=surfname_map, attribute=surfattr_map, realization=int(calc_map)
            )
        if calc_map2 in ["Mean", "StdDev", "Max", "Min", "P90", "P10"]:
            surface2 = surface_set_models[ensemble_map2].calculate_statistical_surface(
                name=surfname_map2,
                attribute=surfattr_map2,
                calculation=calc_map2,
                realizations=realizations,
            )
        else:
            surface2 = surface_set_models[ensemble_map2].get_realization_surface(
                name=surfname_map2, attribute=surfattr_map2, realization=int(calc_map2)
            )

        # Generate Leaflet layers
        update_controls = check_if_update_needed(
            ctx=ctx,
            current_maps=[current_map, current_map2],
            compute_diff=compute_diff,
            color_range_settings=color_range_settings,
        )

        surface_layers = create_or_return_base_layer(
            update_controls,
            surface,
            current_map,
            shade_map,
            color_range_settings,
            map_id="map1",
        )
        surface_layers2 = create_or_return_base_layer(
            update_controls,
            surface2,
            current_map2,
            shade_map2,
            color_range_settings,
            map_id="map2",
        )

        try:
            surface3 = surface.copy()
            surface3.values = surface3.values - surface2.values

            diff_layers = (
                [
                    SurfaceLeafletModel(
                        surface3,
                        name="surface3",
                        apply_shading=shade_map3.get("value", False),
                    ).layer
                ]
                if update_controls["diff_map"]["update"]
                else []
            )
        except ValueError:
            diff_layers = []

        if wellname is not None:
            surface_layers.append(well_layer)
            surface_layers2.append(well_layer)
        if polyline is not None:
            surface_layers.append(poly_layer)
        if xline is not None and source == "xline":
            surface_layers.append(xline_layer)
        if yline is not None and source == "yline":
            surface_layers.append(yline_layer)
        if well_set_model is not None:
            if options is not None or options2 is not None:
                if "intersect_well" in options or "intersect_well" in options2:
                    ### This is potentially a heavy task as it loads all wells into memory
                    wells: List[xtgeo.Well] = list(well_set_model.wells.values())
                if "intersect_well" in options and update_controls["map1"]["update"]:
                    surface_layers.append(
                        create_leaflet_well_marker_layer(wells, surface)
                    )
                if "intersect_well" in options2 and update_controls["map2"]["update"]:
                    surface_layers2.append(
                        create_leaflet_well_marker_layer(wells, surface2)
                    )

        return (
            f"Surface A: {surfattr_map} - {surfname_map} - {ensemble_map} - {calc_map}",
            surface_layers if update_controls["map1"]["update"] else no_update,
            f"Surface B: {surfattr_map2} - {surfname_map2} - {ensemble_map2} - {calc_map2}",
            surface_layers2 if update_controls["map2"]["update"] else no_update,
            "Surface A-B",
            diff_layers if update_controls["diff_map"]["update"] else no_update,
        )

    @app.callback(
        Output({"id": get_uuid("map"), "element": "stored_polyline"}, "data"),
        Input(get_uuid("leaflet-map1"), "polyline_points"),
    )
    def _store_polyline_points(
        positions_yx: List[List[float]],
    ) -> Optional[List[List[float]]]:
        """Stores drawn in polyline in a dcc.Store. Reversing elements to reflect
        normal behaviour"""
        if positions_yx is not None:
            try:
                return [[pos[1], pos[0]] for pos in positions_yx]
            except TypeError:
                warnings.warn("Polyline for map is not valid format")
                return None
        raise PreventUpdate

    @app.callback(
        Output(
            {"id": get_uuid("intersection-data"), "element": "source"},
            "value",
        ),
        Output(
            {"id": get_uuid("intersection-data"), "element": "well"},
            "value",
        ),
        Input(get_uuid("leaflet-map1"), "clicked_shape"),
        Input(get_uuid("leaflet-map1"), "polyline_points"),
    )
    def _update_from_map_click(
        clicked_shape: Optional[Dict],
        _polyline: List[List[float]],
    ) -> Tuple[str, Union[NoUpdate, str]]:
        """Update intersection source and optionally selected well when
        user clicks a shape in map"""
        ctx = callback_context.triggered[0]
        if "polyline_points" in ctx["prop_id"]:
            return "polyline", no_update
        if clicked_shape is None:
            raise PreventUpdate
        if clicked_shape.get("id") == "random_line":
            return "polyline", no_update
        if clicked_shape.get("id") in well_set_model.well_names:
            return "well", clicked_shape.get("id")
        raise PreventUpdate

    @app.callback(
        Output(get_uuid("map-color-ranges"), "data"),
        Output(
            {"id": get_uuid("map-settings"), "colors": "map2_clip_min"},
            "disabled",
        ),
        Output(
            {"id": get_uuid("map-settings"), "colors": "map2_clip_max"},
            "disabled",
        ),
        Input(
            {"id": get_uuid("map-settings"), "colors": "map1_clip_min"},
            "value",
        ),
        Input(
            {"id": get_uuid("map-settings"), "colors": "map1_clip_max"},
            "value",
        ),
        Input(
            {"id": get_uuid("map-settings"), "colors": "map2_clip_min"},
            "value",
        ),
        Input(
            {"id": get_uuid("map-settings"), "colors": "map2_clip_max"},
            "value",
        ),
        Input(
            {"id": get_uuid("map-settings"), "colors": "sync_range"},
            "value",
        ),
    )
    def _color_range_options(
        clip_min_map1: Optional[float],
        clip_max_map1: Optional[float],
        clip_min_map2: Optional[float],
        clip_max_map2: Optional[float],
        sync_range: list,
    ) -> Tuple[Dict[str, Dict], bool, bool]:
        ctx = callback_context.triggered[0]

        return (
            {
                "map1": {
                    "color_range": [clip_min_map1, clip_max_map1],
                    "update": "map1" in ctx["prop_id"],
                },
                "map2": {
                    "color_range": [clip_min_map2, clip_max_map2]
                    if not sync_range
                    else [clip_min_map1, clip_max_map1],
                    "update": "map2" in ctx["prop_id"]
                    or (sync_range and "map1" in ctx["prop_id"])
                    or (
                        "sync_range" in ctx["prop_id"]
                        and [clip_min_map1, clip_max_map1]
                        != [clip_min_map2, clip_max_map2]
                    ),
                },
            },
            bool(sync_range),
            bool(sync_range),
        )


def create_leaflet_polyline_layer(
    positions: List[List[float]], name: str, poly_id: str
) -> Dict:
    return {
        "id": name,
        "name": name,
        "baseLayer": False,
        "checked": True,
        "action": "update",
        "data": [
            {
                "type": "polyline",
                "id": poly_id,
                "positions": positions,
                "color": "blue",
                "tooltip": "polyline",
            },
            {
                "type": "circle",
                "center": positions[0],
                "radius": 60,
                "color": "blue",
                "tooltip": "B",
            },
            {
                "type": "circle",
                "center": positions[-1],
                "radius": 60,
                "color": "blue",
                "tooltip": "B'",
            },
        ],
    }


def replace_or_add_map_layer(
    layers: List[Dict], uuid: str, new_layer: Dict
) -> List[Dict]:
    for idx, layer in enumerate(layers):
        if layer.get("id") == uuid:
            layers[idx] = new_layer
            return layers
    layers.append(new_layer)
    return layers


def check_if_update_needed(
    ctx: Dict,
    current_maps: List[List[Dict]],
    compute_diff: List,
    color_range_settings: Dict,
) -> Dict[str, Any]:

    update_controls = {}
    for map_id, current_map in zip(["map1", "map2"], current_maps):
        map_controllers_clicked = f'"map_id":"{map_id}"' in ctx["prop_id"]
        change_calculate_well_intersections = (
            map_controllers_clicked and "options" in ctx["prop_id"]
        )
        change_shade_map = (
            f"leaflet-{map_id}" in ctx["prop_id"] and "switch" in ctx["prop_id"]
        )
        change_color_clip = (
            "map-color-ranges" in ctx["prop_id"]
            and color_range_settings[map_id]["update"]
        )
        initial_loading = not current_map or all(
            layer.get("id") != map_id for layer in current_map
        )
        update_controls[map_id] = {
            "update": (
                map_controllers_clicked
                or change_shade_map
                or change_color_clip
                or initial_loading
            ),
            "use_base_layer": (change_shade_map or change_calculate_well_intersections),
        }

    change_diffmap_from_options = (
        "leaflet-map3" in ctx["prop_id"] and "switch" in ctx["prop_id"]
    ) or "compute_diff" in ctx["prop_id"]

    update_controls["diff_map"] = {
        "update": compute_diff
        and (
            (
                update_controls["map1"]["update"]
                and not update_controls["map1"]["use_base_layer"]
            )
            or (
                update_controls["map2"]["update"]
                and not update_controls["map2"]["use_base_layer"]
            )
            or change_diffmap_from_options
        )
        and not "map-color-ranges" in ctx["prop_id"]
    }

    return update_controls


def create_or_return_base_layer(
    update_controls: Dict,
    surface: xtgeo.RegularSurface,
    current_map: List[Dict],
    shade_map: Dict[str, bool],
    color_range_settings: Dict,
    map_id: str,
) -> List[Dict]:

    surface_layers = []
    if update_controls[map_id]["use_base_layer"]:
        for layer in current_map:
            if layer["baseLayer"]:
                layer["data"][0]["shader"]["applyHillshading"] = shade_map.get("value")
                surface_layers = [layer]
    else:
        surface_layers = [
            SurfaceLeafletModel(
                surface,
                clip_min=color_range_settings[map_id]["color_range"][0],
                clip_max=color_range_settings[map_id]["color_range"][1],
                name=map_id,
                apply_shading=shade_map.get("value", False),
            ).layer
        ]
    return surface_layers



from typing import Callable, Dict, List, Optional

import webviz_core_components as wcc
from dash import dcc, html

from .dialog import open_dialog_layout


def intersection_data_layout(
    get_uuid: Callable,
    surface_attributes: List[str],
    surface_names: List[str],
    ensembles: List[str],
    use_wells: bool,
    well_names: List[str],
    surface_geometry: Dict,
    initial_settings: Dict,
) -> html.Div:
    """Layout for selecting intersection data"""
    return html.Div(
        id=get_uuid("intersection-data-wrapper"),
        children=[
            html.Div(
                style={
                    "padding-bottom": "10px",
                    "border-bottom-style": "solid",
                    "border-width": "thin",
                    "border-color": "grey",
                },
                id=get_uuid("intersection-source-wrapper"),
                children=[
                    source_layout(
                        uuid=get_uuid("intersection-data"),
                        use_wells=use_wells,
                    ),
                    well_layout(
                        uuid=get_uuid("intersection-data"),
                        well_names=well_names,
                        value=initial_settings.get(
                            "well",
                            well_names[0] if use_wells else None,
                        ),
                    ),
                    xline_layout(
                        uuid=get_uuid("intersection-data"),
                        surface_geometry=surface_geometry,
                    ),
                    yline_layout(
                        uuid=get_uuid("intersection-data"),
                        surface_geometry=surface_geometry,
                    ),
                ],
            ),
            surface_attribute_layout(
                uuid=get_uuid("intersection-data"),
                surface_attributes=surface_attributes,
                value=initial_settings.get(
                    "surface_attribute",
                    surface_attributes[0],
                ),
            ),
            surface_names_layout(
                uuid=get_uuid("intersection-data"),
                surface_names=surface_names,
                value=initial_settings.get("surface_names", [surface_names[0]]),
            ),
            ensemble_layout(
                uuid=get_uuid("intersection-data"),
                ensemble_names=ensembles,
                value=initial_settings.get("ensembles", [ensembles[0]]),
            ),
            statistical_layout(
                uuid=get_uuid("intersection-data"),
                value=initial_settings.get("calculation", ["Mean", "Min", "Max"]),
            ),
            blue_apply_button(
                uuid=get_uuid("apply-intersection-data-selections"),
                title="Update intersection",
            ),
            open_dialog_layout(
                dialog_id="uncertainty-table",
                uuid=get_uuid("dialog"),
                title="Uncertainty table",
            ),
            settings_layout(get_uuid, initial_settings=initial_settings),
        ],
    )


def source_layout(uuid: str, use_wells: bool = True) -> html.Div:
    options = [
        {"label": "Intersect polyline from Surface A", "value": "polyline"},
        {"label": "Intersect x-line from Surface A", "value": "xline"},
        {"label": "Intersect y-line from Surface A", "value": "yline"},
    ]
    if use_wells:
        options.append({"label": "Intersect well", "value": "well"})
    return wcc.Dropdown(
        label="Intersection source",
        id={"id": uuid, "element": "source"},
        options=options,
        value="well" if use_wells else "polyline",
        clearable=False,
    )


def well_layout(
    uuid: str, well_names: List[str], value: Optional[str] = None
) -> html.Div:
    return html.Div(
        style={
            "display": "none",
        }
        if value is None
        else {},
        id={"id": uuid, "element": "well-wrapper"},
        children=wcc.Dropdown(
            label="Well",
            id={"id": uuid, "element": "well"},
            options=[{"label": well, "value": well} for well in well_names],
            value=value,
            clearable=False,
        ),
    )


def xline_layout(uuid: str, surface_geometry: Dict) -> html.Div:
    return html.Div(
        style={
            "display": "none",
        },
        id={"id": uuid, "element": "xline-wrapper"},
        children=[
            html.Label("X-Line:"),
            wcc.FlexBox(
                style={"fontSize": "0.8em"},
                children=[
                    dcc.Input(
                        id={"id": uuid, "cross-section": "xline", "element": "value"},
                        style={"flex": 3, "minWidth": "100px"},
                        type="number",
                        value=round(surface_geometry["xmin"]),
                        min=round(surface_geometry["xmin"]),
                        max=round(surface_geometry["xmax"]),
                        step=500,
                        persistence=True,
                        persistence_type="session",
                    ),
                    wcc.Label(
                        style={"flex": 1, "marginLeft": "10px", "minWidth": "20px"},
                        children="Step:",
                    ),
                    dcc.Input(
                        id={"id": uuid, "cross-section": "xline", "element": "step"},
                        style={"flex": 2, "minWidth": "20px"},
                        value=500,
                        type="number",
                        min=1,
                        max=round(surface_geometry["xmax"])
                        - round(surface_geometry["xmin"]),
                        persistence=True,
                        persistence_type="session",
                    ),
                ],
            ),
        ],
    )


def yline_layout(uuid: str, surface_geometry: Dict) -> html.Div:
    return html.Div(
        style={
            "display": "none",
        },
        id={"id": uuid, "element": "yline-wrapper"},
        children=[
            html.Label("Y-Line:"),
            wcc.FlexBox(
                style={"fontSize": "0.8em"},
                children=[
                    dcc.Input(
                        id={"id": uuid, "cross-section": "yline", "element": "value"},
                        style={"flex": 3, "minWidth": "100px"},
                        type="number",
                        value=round(surface_geometry["ymin"]),
                        min=round(surface_geometry["ymin"]),
                        max=round(surface_geometry["ymax"]),
                        step=50,
                        persistence=True,
                        persistence_type="session",
                    ),
                    wcc.Label(
                        style={"flex": 1, "marginLeft": "10px", "minWidth": "20px"},
                        children="Step:",
                    ),
                    dcc.Input(
                        id={"id": uuid, "cross-section": "yline", "element": "step"},
                        style={"flex": 2, "minWidth": "20px"},
                        value=50,
                        type="number",
                        min=1,
                        max=round(surface_geometry["ymax"])
                        - round(surface_geometry["ymin"]),
                        persistence=True,
                        persistence_type="session",
                    ),
                ],
            ),
        ],
    )


def surface_attribute_layout(
    uuid: str, surface_attributes: List[str], value: str
) -> html.Div:
    return wcc.Dropdown(
        label="Surface attribute",
        id={"id": uuid, "element": "surface_attribute"},
        options=[
            {"label": attribute, "value": attribute} for attribute in surface_attributes
        ],
        value=value,
        clearable=False,
        multi=False,
    )


def surface_names_layout(
    uuid: str, surface_names: List[str], value: List[str]
) -> html.Div:
    return wcc.SelectWithLabel(
        label="Surface names",
        id={"id": uuid, "element": "surface_names"},
        options=[
            {"label": attribute, "value": attribute} for attribute in surface_names
        ],
        value=value,
        multi=True,
        size=min(len(surface_names), 5),
    )


def ensemble_layout(uuid: str, ensemble_names: List[str], value: List[str]) -> html.Div:
    return html.Div(
        style={
            "marginTop": "5px",
            "display": ("inline" if len(ensemble_names) > 1 else "none"),
        },
        children=wcc.SelectWithLabel(
            label="Ensembles",
            id={"id": uuid, "element": "ensembles"},
            options=[{"label": ens, "value": ens} for ens in ensemble_names],
            value=value,
            size=min(len(ensemble_names), 3),
        ),
    )


def statistical_layout(uuid: str, value: List[str]) -> html.Div:
    return wcc.Checklist(
        label="Show surfaces",
        id={"id": uuid, "element": "calculation"},
        options=[
            {"label": "Mean", "value": "Mean"},
            {"label": "Min", "value": "Min"},
            {"label": "Max", "value": "Max"},
            {"label": "Realizations", "value": "Realizations"},
            {
                "label": "Uncertainty envelope",
                "value": "Uncertainty envelope",
            },
        ],
        value=value,
    )


def blue_apply_button(uuid: str, title: str) -> html.Div:
    return html.Button(
        title,
        className="webviz-structunc-blue-apply-btn",
        id=uuid,
    )


def options_layout(
    uuid: str,
    depth_truncations: Dict,
    resolution: float,
    extension: int,
    initial_layout: Dict,
) -> html.Div:

    return html.Div(
        children=[
            html.Div(
                children=[
                    wcc.Label(
                        "Resolution (m) ",
                    ),
                    dcc.Input(
                        className="webviz-structunc-range-input",
                        id={"id": uuid, "element": "resolution"},
                        type="number",
                        required=True,
                        value=resolution,
                        persistence=True,
                        persistence_type="session",
                    ),
                ],
            ),
            html.Div(
                children=[
                    wcc.Label(
                        "Extension (m) ",
                    ),
                    dcc.Input(
                        className="webviz-structunc-range-input",
                        id={"id": uuid, "element": "extension"},
                        type="number",
                        step=25,
                        required=True,
                        value=extension,
                        persistence=True,
                        persistence_type="session",
                    ),
                ],
            ),
            html.Div(
                style={"margin-top": "10px"},
                children=[
                    wcc.Label("Depth range settings:"),
                    wcc.FlexBox(
                        style={"display": "flex"},
                        children=[
                            dcc.Input(
                                id={
                                    "id": uuid,
                                    "settings": "zrange_min",
                                },
                                style={"flex": 1, "minWidth": "70px"},
                                type="number",
                                value=depth_truncations.get("min", None),
                                debounce=True,
                                placeholder="Min",
                                persistence=True,
                                persistence_type="session",
                            ),
                            dcc.Input(
                                id={
                                    "id": uuid,
                                    "settings": "zrange_max",
                                },
                                style={"flex": 1, "minWidth": "70px"},
                                type="number",
                                value=depth_truncations.get("max", None),
                                debounce=True,
                                placeholder="Max",
                                persistence=True,
                                persistence_type="session",
                            ),
                        ],
                    ),
                    wcc.RadioItems(
                        id={
                            "id": uuid,
                            "settings": "zrange_locks",
                        },
                        options=[
                            {
                                "label": "Truncate range",
                                "value": "truncate",
                            },
                            {
                                "label": "Lock range",
                                "value": "lock",
                            },
                        ],
                        value="truncate",
                    ),
                    wcc.Checklist(
                        id={
                            "id": uuid,
                            "settings": "ui_options",
                        },
                        options=[
                            {"label": "Keep zoom state", "value": "uirevision"},
                        ],
                        value=["uirevision"] if "uirevision" in initial_layout else [],
                    ),
                ],
            ),
        ],
    )


def settings_layout(get_uuid: Callable, initial_settings: Dict) -> html.Div:
    return wcc.Selectors(
        open_details=False,
        label="⚙️ Settings",
        children=[
            options_layout(
                uuid=get_uuid("intersection-data"),
                depth_truncations=initial_settings.get("depth_truncations", {}),
                resolution=initial_settings.get("resolution", 10),
                extension=initial_settings.get("extension", 500),
                initial_layout=initial_settings.get("intersection_layout", {}),
            ),
            open_dialog_layout(
                dialog_id="color",
                uuid=get_uuid("dialog"),
                title="Intersection colors",
            ),
        ],
    )



import numpy as np

from webviz_subsurface._datainput.image_processing import array_to_png

with open("tests/data/surface_png.txt", "r") as file:
    BASE64_SURFACE = file.read()


def test_array_to_png() -> None:
    data = np.loadtxt("tests/data/surface_zarr.np.gz")
    assert array_to_png(data) == BASE64_SURFACE



from typing import Optional

import plotly.graph_objects as go
import webviz_core_components as wcc
from dash import dash_table, dcc, html

from webviz_subsurface._models import InplaceVolumesModel


def comparison_main_layout(uuid: str) -> html.Div:
    return [
        html.Div(
            style={"margin-bottom": "20px"},
            children=wcc.RadioItems(
                vertical=False,
                id={"id": uuid, "element": "display-option"},
                options=[
                    {
                        "label": "QC plots",
                        "value": "plots",
                    },
                    {
                        "label": "Difference table for selected response",
                        "value": "single-response table",
                    },
                    {
                        "label": "Difference table for multiple responses",
                        "value": "multi-response table",
                    },
                ],
                value="plots",
            ),
        ),
        html.Div(id={"id": uuid, "wrapper": "table"}),
    ]


def comparison_qc_plots_layout(
    fig_diff_vs_real: Optional[go.Figure],
    fig_corr: go.Figure,
    fig_diff_vs_response: go.Figure,
    barfig: go.Figure,
) -> html.Div:
    real_plot = fig_diff_vs_real is not None
    return html.Div(
        children=[
            html.Div(
                children=wcc.Graph(
                    config={"displayModeBar": False},
                    style={"height": "21vh"},
                    figure=fig_diff_vs_real,
                )
                if real_plot
                else [],
            ),
            wcc.FlexBox(
                style={"height": "31vh" if real_plot else "52vh"},
                children=[
                    wcc.FlexColumn(
                        children=wcc.Graph(
                            config={"displayModeBar": False},
                            style={"height": "30vh" if real_plot else "51vh"},
                            figure=fig_diff_vs_response,
                        )
                    ),
                    wcc.FlexColumn(
                        children=wcc.Graph(
                            config={"displayModeBar": False},
                            style={"height": "30vh" if real_plot else "51vh"},
                            figure=fig_corr,
                        )
                    ),
                ],
            ),
            wcc.Frame(
                style={"height": "29vh"},
                children=[
                    wcc.Header("Highlighted data"),
                    wcc.Graph(
                        config={"displayModeBar": False},
                        style={"height": "25vh"},
                        figure=barfig,
                    )
                    if barfig is not None
                    else html.Div("No data within highlight criteria"),
                ],
            ),
        ]
    )


def comparison_table_layout(
    table: dash_table, table_type: str, selections: dict, filter_info: str
) -> html.Div:
    if table_type == "single-response table":
        header = f"Table showing differences for {selections['Response']}"
    else:
        diff_mode = "percent" if selections["Diff mode"] == "diff (%)" else "true value"
        header = f"Table showing differences in {diff_mode} for multiple responses"

    return html.Div(
        children=[
            wcc.Header(header),
            html.Div(
                style={"margin-bottom": "30px", "font-weight": "bold"},
                children=[
                    html.Div(
                        f"From {selections['value1'].replace('|', ':  ')} "
                        f"to {selections['value2'].replace('|', ':  ')}"
                    ),
                    html.Div(
                        f"{filter_info.capitalize()} {selections['filters'][filter_info][0]}"
                    ),
                ],
            ),
            table,
        ]
    )


def comparison_options(compare_on: str, volumemodel: InplaceVolumesModel) -> list:
    if compare_on == "Ensemble" and "SENSNAME_CASE" in volumemodel.selectors:
        elements = []
        for ens in volumemodel.ensembles:
            for sens in volumemodel.ensemble_sensitivities[ens]:
                elements.append("|".join([ens, sens]))
        return [{"label": i.replace("|", ":  "), "value": i} for i in elements]

    if compare_on == "Source":
        elements = volumemodel.sources
    elif compare_on == "Ensemble":
        elements = volumemodel.ensembles
    elif compare_on == "Sensitivity":
        elements = list(volumemodel.ensemble_sensitivities.values())[0]
    return [{"label": i, "value": i} for i in elements]


def comparison_selections(
    uuid: str, volumemodel: InplaceVolumesModel, tab: str, compare_on: str
) -> html.Div:
    options = comparison_options(compare_on, volumemodel)
    return html.Div(
        children=[
            wcc.Selectors(
                label="CONTROLS",
                open_details=True,
                children=[
                    wcc.Dropdown(
                        id={"id": uuid, "tab": tab, "selector": "value1"},
                        label=f"{compare_on} A",
                        optionHeight=50,
                        options=options,
                        value=options[0]["value"],
                        clearable=False,
                    ),
                    wcc.Dropdown(
                        id={"id": uuid, "tab": tab, "selector": "value2"},
                        optionHeight=50,
                        label=f"{compare_on} B",
                        options=options,
                        value=options[-1]["value"],
                        clearable=False,
                    ),
                    dcc.Input(
                        style={"display": "none"},
                        id={"id": uuid, "tab": tab, "selector": "compare_on"},
                        value=compare_on,
                    ),
                    html.Div(
                        f"Difference = {compare_on} B - {compare_on} A",
                        style={
                            "font-size": "15px",
                            "margin-top": "5px",
                            "color": "#007079",
                        },
                    ),
                    response_selector(volumemodel, uuid, tab),
                    group_by_selector(volumemodel, uuid, tab),
                    diff_mode_selector(uuid, tab),
                    highlight_controls(uuid, tab),
                ],
            ),
            wcc.Selectors(
                label="⚙️ SETTINGS",
                open_details=False,
                children=[
                    colorby_selector(uuid, tab),
                    axis_focus_selector(uuid, tab),
                    remove_zero_responses(uuid, tab),
                    remove_non_highlighted_data(uuid, tab),
                ],
            ),
        ]
    )


def axis_focus_selector(uuid: str, tab: str) -> html.Div:
    return wcc.Checklist(
        id={"id": uuid, "tab": tab, "selector": "Axis focus"},
        options=[{"label": "Focus diff plots on highlighted", "value": "focus"}],
        value=["focus"],
    )


def remove_zero_responses(uuid: str, tab: str) -> html.Div:
    return wcc.Checklist(
        id={"id": uuid, "tab": tab, "selector": "Remove zeros"},
        options=[{"label": "Remove data with no volume", "value": "remove"}],
        value=["remove"],
    )


def remove_non_highlighted_data(uuid: str, tab: str) -> html.Div:
    return wcc.Checklist(
        id={"id": uuid, "tab": tab, "selector": "Remove non-highlighted"},
        options=[
            {"label": "Display only highlighted data in table", "value": "remove"}
        ],
        value=[],
    )


def diff_mode_selector(uuid: str, tab: str) -> html.Div:
    return html.Div(
        style={"margin-top": "10px"},
        children=wcc.RadioItems(
            label="Difference mode",
            id={"id": uuid, "tab": tab, "selector": "Diff mode"},
            options=[
                {"label": "Percent", "value": "diff (%)"},
                {"label": "True value", "value": "diff"},
            ],
            labelStyle={"display": "inline-flex", "margin-right": "5px"},
            value="diff (%)",
        ),
    )


def highlight_controls(uuid: str, tab: str) -> html.Div:
    return html.Div(
        style={"margin-top": "10px"},
        children=[
            html.Label("Data highlight criterias", className="webviz-underlined-label"),
            html.Div(
                children=[
                    wcc.Label("Absolute diff (%) above:"),
                    dcc.Input(
                        id={"id": uuid, "tab": tab, "selector": "Accept value"},
                        type="number",
                        required=True,
                        value=5,
                        persistence=True,
                        persistence_type="session",
                    ),
                ],
            ),
            html.Div(
                children=[
                    wcc.Label("Ignore response values below:"),
                    dcc.Input(
                        id={"id": uuid, "tab": tab, "selector": "Ignore <"},
                        type="number",
                        required=True,
                        value=0,
                        persistence=True,
                        persistence_type="session",
                        debounce=True,
                    ),
                ]
            ),
        ],
    )


def response_selector(
    volumemodel: InplaceVolumesModel,
    uuid: str,
    tab: str,
) -> html.Div:
    return html.Div(
        style={"margin-top": "10px"},
        children=wcc.Dropdown(
            id={
                "id": uuid,
                "tab": tab,
                "selector": "Response",
            },
            label="Response",
            options=[{"label": i, "value": i} for i in volumemodel.responses],
            value=volumemodel.volume_columns[0],
            clearable=False,
        ),
    )


def colorby_selector(
    uuid: str,
    tab: str,
) -> html.Div:
    return html.Div(
        style={"margin": "10px 0px"},
        children=wcc.RadioItems(
            label="Color plots on",
            id={"id": uuid, "tab": tab, "selector": "Color by"},
            options=[
                {"label": "Highlighted", "value": "highlighted"},
                {"label": "1st level of investigation", "value": "groups"},
            ],
            labelStyle={"display": "inline-flex", "margin-right": "5px"},
            value="highlighted",
        ),
    )


def group_by_selector(
    volumemodel: InplaceVolumesModel, uuid: str, tab: str
) -> html.Div:
    available_selectors = [
        x
        for x in volumemodel.region_selectors + ["FACIES", "REAL", "FLUID_ZONE"]
        if x in volumemodel.selectors and volumemodel.dataframe[x].nunique() > 1
    ]
    return html.Div(
        style={"margin-top": "10px"},
        children=wcc.Dropdown(
            label="Investigate differences on level",
            id={"id": uuid, "tab": tab, "selector": "Group by"},
            options=[{"label": elm, "value": elm} for elm in available_selectors],
            value=[],
            placeholder="Total",
            multi=True,
            clearable=False,
        ),
    )



# pylint: disable=no-name-in-module
from webviz_config.plugins import ParameterDistribution


def test_parameter_distribution(dash_duo, app, shared_settings) -> None:
    plugin = ParameterDistribution(
        app,
        shared_settings["HM_SETTINGS"],
        ensembles=shared_settings["HM_ENSEMBLES"],
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



import hashlib
import logging
import os
from pathlib import Path
from typing import Optional

from webviz_config.webviz_factory import WebvizFactory
from webviz_config.webviz_factory_registry import WEBVIZ_FACTORY_REGISTRY
from webviz_config.webviz_instance_info import WebvizRunMode

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._arrow_unsmry_import import load_per_realization_arrow_unsmry_files
from ._csv_import import (
    load_ensemble_summary_csv_file,
    load_per_real_csv_file_using_fmu,
)
from ._provider_impl_arrow_lazy import ProviderImplArrowLazy
from ._provider_impl_arrow_presampled import ProviderImplArrowPresampled
from ._resampling import Frequency, resample_single_real_table
from .ensemble_summary_provider import EnsembleSummaryProvider

LOGGER = logging.getLogger(__name__)


class EnsembleSummaryProviderFactory(WebvizFactory):
    def __init__(self, root_storage_folder: Path, allow_storage_writes: bool) -> None:
        self._storage_dir = Path(root_storage_folder) / __name__
        self._allow_storage_writes = allow_storage_writes

        LOGGER.info(
            f"EnsembleSummaryProviderFactory init: storage_dir={self._storage_dir}"
        )

        if self._allow_storage_writes:
            os.makedirs(self._storage_dir, exist_ok=True)

    @staticmethod
    def instance() -> "EnsembleSummaryProviderFactory":
        """Static method to access the singleton instance of the factory."""

        factory = WEBVIZ_FACTORY_REGISTRY.get_factory(EnsembleSummaryProviderFactory)
        if not factory:
            app_instance_info = WEBVIZ_FACTORY_REGISTRY.app_instance_info
            storage_folder = app_instance_info.storage_folder
            allow_writes = app_instance_info.run_mode != WebvizRunMode.PORTABLE

            factory = EnsembleSummaryProviderFactory(storage_folder, allow_writes)

            # Store the factory object in the global factory registry
            WEBVIZ_FACTORY_REGISTRY.set_factory(EnsembleSummaryProviderFactory, factory)

        return factory

    def create_from_ensemble_csv_file(
        self,
        csv_file: Path,
        ensemble_filter: Optional[str] = None,
    ) -> EnsembleSummaryProvider:
        """Create EnsembleSummaryProvider from aggregated CSV file.
        The CSV file is assumed to contain data for a single ensemble and must contain
        columns for `REAL` and `DATE` in addition to the actual numeric vectors.
        If the CSV file contains an `ENSEMBLE` column it will be ignored, but an exception
        will be thrown if it is present and it contains multiple ensemble names.

        Note that the returned summary provider does not support resampling, nor will it
        be able to return vector metadata.
        """

        timer = PerfTimer()

        storage_key = "ens_csv"
        if ensemble_filter is not None:
            storage_key += f"_filtered_on_{ensemble_filter}"
        storage_key += f"__{_make_hash_string(str(csv_file))}"

        provider = ProviderImplArrowPresampled.from_backing_store(
            self._storage_dir, storage_key
        )
        if provider:
            LOGGER.info(
                f"Loaded summary provider (CSV) from backing store in "
                f"{timer.elapsed_s():.2f}s (csv_file={csv_file})"
            )
            return provider

        # We can only import data from CSV if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(f"Failed to load summary provider (CSV) for {csv_file}")

        LOGGER.info(f"Importing/saving CSV summary data for: {csv_file}")

        timer.lap_s()
        ensemble_df = load_ensemble_summary_csv_file(csv_file, ensemble_filter)
        et_import_csv_s = timer.lap_s()

        if len(ensemble_df) == 0:
            raise ValueError("Import resulted in empty DataFrame")
        if "DATE" not in ensemble_df.columns:
            raise ValueError("No DATE column present in input data")
        if "REAL" not in ensemble_df.columns:
            raise ValueError("No REAL column present in input data")

        ProviderImplArrowPresampled.write_backing_store_from_ensemble_dataframe(
            self._storage_dir, storage_key, ensemble_df
        )
        et_write_s = timer.lap_s()

        provider = ProviderImplArrowPresampled.from_backing_store(
            self._storage_dir, storage_key
        )
        if not provider:
            raise ValueError(f"Failed to load/create provider for {csv_file}")

        LOGGER.info(
            f"Saved summary provider (CSV) to backing store in {timer.elapsed_s():.2f}s ("
            f"import_csv={et_import_csv_s:.2f}s, "
            f"write={et_write_s:.2f}s, "
            f"csv_file={csv_file})"
        )

        return provider

    def create_from_per_realization_csv_file(
        self, ens_path: str, csv_file_rel_path: str
    ) -> EnsembleSummaryProvider:
        """Create EnsembleSummaryProvider from per realization CSV files.

        Note that the returned summary provider does not support resampling, nor will it
        be able to return vector metadata.
        """

        timer = PerfTimer()

        storage_key = f"per_real_csv__{_make_hash_string(ens_path + csv_file_rel_path)}"
        provider = ProviderImplArrowPresampled.from_backing_store(
            self._storage_dir, storage_key
        )

        if provider:
            LOGGER.info(
                f"Loaded summary provider (per real CSV) from backing store in "
                f"{timer.elapsed_s():.2f}s ("
                f"ens_path={ens_path}, csv_file_rel_path={csv_file_rel_path})"
            )
            return provider

        # We can only import data from CSV if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(
                f"Failed to load summary provider (per real CSV) for {ens_path}"
            )

        LOGGER.info(f"Importing/saving per real CSV summary data for: {ens_path}")

        timer.lap_s()

        ensemble_df = load_per_real_csv_file_using_fmu(ens_path, csv_file_rel_path)
        et_import_csv_s = timer.lap_s()

        ProviderImplArrowPresampled.write_backing_store_from_ensemble_dataframe(
            self._storage_dir, storage_key, ensemble_df
        )
        et_write_s = timer.lap_s()

        provider = ProviderImplArrowPresampled.from_backing_store(
            self._storage_dir, storage_key
        )

        if not provider:
            raise ValueError(
                f"Failed to load/create provider (per real CSV) for {ens_path}"
            )

        LOGGER.info(
            f"Saved summary provider (per real CSV) to backing store in {timer.elapsed_s():.2f}s ("
            f"import_csv={et_import_csv_s:.2f}s, write={et_write_s:.2f}s, "
            f"ens_path={ens_path}, csv_file_rel_path={csv_file_rel_path})"
        )

        return provider

    def create_from_arrow_unsmry_lazy(
        self, ens_path: str, rel_file_pattern: str
    ) -> EnsembleSummaryProvider:
        """Create EnsembleSummaryProvider from per-realization unsmry data in .arrow format.

        The `rel_file_pattern` parameter must specify a relative (per realization) file pattern
        that will be used to find the wanted .arrow files within each realization. The file
        pattern is relative to each realization's `runpath`.
        Typically the file pattern will be: "share/results/unsmry/*.arrow"

        The returned summary provider supports lazy resampling.
        """

        timer = PerfTimer()

        storage_key = (
            f"arrow_unsmry_lazy__{_make_hash_string(ens_path + rel_file_pattern)}"
        )
        provider = ProviderImplArrowLazy.from_backing_store(
            self._storage_dir, storage_key
        )
        if provider:
            LOGGER.info(
                f"Loaded lazy summary provider from backing store in {timer.elapsed_s():.2f}s ("
                f"ens_path={ens_path})"
            )
            return provider

        # We can only import data from data source if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(f"Failed to load lazy summary provider for {ens_path}")

        LOGGER.info(f"Importing/saving arrow summary data for: {ens_path}")

        timer.lap_s()
        per_real_tables = load_per_realization_arrow_unsmry_files(
            ens_path, rel_file_pattern
        )
        if not per_real_tables:
            raise ValueError(
                f"Could not find any .arrow unsmry files for ens_path={ens_path}"
            )
        et_import_smry_s = timer.lap_s()

        try:
            ProviderImplArrowLazy.write_backing_store_from_per_realization_tables(
                self._storage_dir, storage_key, per_real_tables
            )
        except ValueError as exc:
            raise ValueError(f"Failed to write backing store for: {ens_path}") from exc

        et_write_s = timer.lap_s()

        provider = ProviderImplArrowLazy.from_backing_store(
            self._storage_dir, storage_key
        )
        if not provider:
            raise ValueError(f"Failed to load/create lazy provider for {ens_path}")

        LOGGER.info(
            f"Saved lazy summary provider to backing store in {timer.elapsed_s():.2f}s ("
            f"import_smry={et_import_smry_s:.2f}s, write={et_write_s:.2f}s, ens_path={ens_path})"
        )

        return provider

    def create_from_arrow_unsmry_presampled(
        self,
        ens_path: str,
        rel_file_pattern: str,
        sampling_frequency: Optional[Frequency],
    ) -> EnsembleSummaryProvider:
        """Create EnsembleSummaryProvider from per-realization unsmry data in .arrow format.

        The `rel_file_pattern` parameter must specify a relative (per realization) file pattern
        that will be used to find the wanted .arrow files within each realization. The file
        pattern is relative to each realization's `runpath`.
        Typically the file pattern will be: "share/results/unsmry/*.arrow"

        This factory method will sample the input data according to the specified
        `sampling_frequency` during import.

        The returned summary provider does not support lazy resampling, but will always
        return data with the above specified frequency .
        """

        timer = PerfTimer()

        freq_str = sampling_frequency.value if sampling_frequency else "raw"
        hash_str = _make_hash_string(ens_path + rel_file_pattern)
        storage_key = f"arrow_unsmry_presampled_{freq_str}__{hash_str}"
        provider = ProviderImplArrowPresampled.from_backing_store(
            self._storage_dir, storage_key
        )
        if provider:
            LOGGER.info(
                f"Loaded presampled summary provider from backing store in "
                f"{timer.elapsed_s():.2f}s ("
                f"sampling_frequency={sampling_frequency}, ens_path={ens_path})"
            )
            return provider

        # We can only import data from data source if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(
                f"Failed to load presampled summary provider for {ens_path}"
            )

        LOGGER.info(f"Importing/saving arrow summary data for: {ens_path}")

        timer.lap_s()
        per_real_tables = load_per_realization_arrow_unsmry_files(
            ens_path, rel_file_pattern
        )
        if not per_real_tables:
            raise ValueError(
                f"Could not find any .arrow unsmry files for ens_path={ens_path}"
            )
        et_import_smry_s = timer.lap_s()

        if sampling_frequency is not None:
            for real_num, table in per_real_tables.items():
                per_real_tables[real_num] = resample_single_real_table(
                    table, sampling_frequency
                )
        et_resample_s = timer.lap_s()

        ProviderImplArrowPresampled.write_backing_store_from_per_realization_tables(
            self._storage_dir, storage_key, per_real_tables
        )
        et_write_s = timer.lap_s()

        provider = ProviderImplArrowPresampled.from_backing_store(
            self._storage_dir, storage_key
        )
        if not provider:
            raise ValueError(f"Failed to load/create provider for {ens_path}")

        LOGGER.info(
            f"Saved presampled summary provider to backing store in {timer.elapsed_s():.2f}s ("
            f"import_smry={et_import_smry_s:.2f}s, "
            f"resample={et_resample_s:.2f}s, "
            f"write={et_write_s:.2f}s, "
            f"ens_path={ens_path})"
        )

        return provider


def _make_hash_string(string_to_hash: str) -> str:
    # There is no security risk here and chances of collision should be very slim
    return hashlib.md5(string_to_hash.encode()).hexdigest()  # nosec



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import numpy as np


@dataclass
class LineData:
    """
    Definition of line trace data for statistics plot

    `Attributes:`
    * `data`  - 1D np.array of value data
    * `name`  - Name of line data
    """

    data: np.ndarray
    name: str


@dataclass
class StatisticsData:
    """
    Dataclass defining statistics data utilized in creation of statistical plot traces

    `Attributes:`
    * `samples` - Common sample point list for each following value list.
    * `free_line` - LineData with name and value data for free line trace in statistics plot
     (e.g. mean, median, etc.)
    * `minimum` - Optional 1D np.array of minimum value data for statistics plot
    * `maximum` - Optional 1D np.array of maximum value data for statistics plot
    * `low` - Optional low percentile, name and 1D np.array data for statistics plot
    * `mid` - Optional middle percentile, name and 1D np.array data for statistics plot
    * `high` - Optional high percentile, name and 1D np.array data for statistics plot

    """

    # TODO:
    # - Rename mid percentile, find better name?
    # - Consider to replace all lines with List[LineData], where each free line must be
    # named and provided data.
    # - Can then be used for individual realization plots as well?
    # - One suggestion: Create base class with: samples: list, free_lines: List[LineData]
    # and inherit for "StatisticsData". Base class can be utilized for realization plots?

    samples: list = field(default_factory=list)
    free_line: Optional[LineData] = None
    minimum: Optional[np.ndarray] = None
    maximum: Optional[np.ndarray] = None
    low: Optional[LineData] = None
    high: Optional[LineData] = None
    mid: Optional[LineData] = None


def validate_statistics_data(data: StatisticsData) -> None:
    """
    Validation of statistics data

    Ensure equal length of all statistical data lists and x-axis data list

    Raise ValueError if lengths are unequal
    """
    samples_length = len(data.samples)
    if samples_length <= 0:
        raise ValueError("Empty x-axis data list in StatisticsData")
    if data.free_line is not None and samples_length != len(data.free_line.data):
        raise ValueError(
            "Invalid statistics mean value data length. len(data.samples) != len(free_line.data)"
        )
    if data.minimum is not None and samples_length != len(data.minimum):
        raise ValueError(
            "Invalid statistics minimum value data length. len(data.samples) "
            "!= len(data.minimum)"
        )
    if data.maximum is not None and samples_length != len(data.maximum):
        raise ValueError(
            "Invalid statistics maximum value data length. len(data.samples) != "
            "len(data.maximum)"
        )
    if data.low is not None and samples_length != len(data.low.data):
        raise ValueError(
            "Invalid statistics low percentile value data length. len(data.samples) "
            "!= len(data.low.data)"
        )
    if data.mid is not None and samples_length != len(data.mid.data):
        raise ValueError(
            "Invalid statistics middle percentile value data length. len(data.samples) "
            "!= len(data.mid.data)"
        )
    if data.high is not None and samples_length != len(data.high.data):
        raise ValueError(
            "Invalid statistics high percentile value data length. "
            "len(data.samples) != len(data.high.data)"
        )


# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
def create_statistics_traces(
    data: StatisticsData,
    color: str,
    legend_group: str,
    legend_name: Optional[str] = None,
    line_shape: str = "linear",
    line_width: int = 2,
    xaxis: str = "x",
    yaxis: str = "y",
    show_legend: bool = True,
    show_hoverinfo: bool = True,
    hovertext: str = "",
    hovertemplate: Optional[str] = None,
    hovermode: Optional[str] = None,
    legendrank: Optional[int] = None,
) -> List[Dict[str, Any]]:
    """
    Utility function for creating statistical plot traces

    Takes `data` containing data for each statistical feature as input, and creates a list of traces
    for each feature. Plotly plots traces from front to end of the list, thereby the last trace is
    plotted on top.

    Note that the data is optional, which implies that only wanted statistical features needs to be
    provided for trace plot generation.

    The function provides a list of traces: [trace0, tract1, ..., traceN]

    Note:
    If hovertemplate is proved it overrides the hovertext

    Returns:
    List of statistical line traces, one for each statistical feature in data input.
    [trace0, tract1, ..., traceN].
    """

    validate_statistics_data(data)

    def get_default_trace(statistics_name: str, values: np.ndarray) -> Dict[str, Any]:
        trace = {
            "name": legend_name if legend_name else legend_group,
            "x": data.samples,
            "y": values,
            "xaxis": xaxis,
            "yaxis": yaxis,
            "mode": "lines",
            "line": {"width": line_width, "color": color, "shape": line_shape},
            "legendgroup": legend_group,
            "showlegend": False,
        }
        if legendrank:
            trace["legendrank"] = legendrank
        if not show_hoverinfo:
            trace["hoverinfo"] = "skip"
            return trace
        if hovertemplate is not None:
            trace["hovertemplate"] = hovertemplate + statistics_name
        else:
            trace["hovertext"] = statistics_name + " " + hovertext
        if hovermode is not None:
            trace["hovermode"] = hovermode
        return trace

    traces: List[Dict[str, Any]] = []

    # Minimum
    if data.minimum is not None:
        minimum_trace = get_default_trace(
            statistics_name="Minimum",
            values=data.minimum,
        )
        minimum_trace["line"] = {
            "color": color,
            "shape": line_shape,
            "dash": "longdash",
            "width": line_width,
        }
        traces.append(minimum_trace)

    # Low percentile
    if data.low is not None:
        low_trace = get_default_trace(
            statistics_name=data.low.name, values=data.low.data
        )
        low_trace["line"] = {
            "width": line_width,
            "color": color,
            "shape": line_shape,
            "dash": "dashdot",
        }
        traces.append(low_trace)

    # Mid percentile
    if data.mid is not None:
        mid_trace = get_default_trace(
            statistics_name=data.mid.name, values=data.mid.data
        )
        mid_trace["line"] = {
            "color": color,
            "shape": line_shape,
            "dash": "dot",
            "width": line_width,
        }
        traces.append(mid_trace)

    # High percentile
    if data.high is not None:
        high_trace = get_default_trace(
            statistics_name=data.high.name, values=data.high.data
        )
        high_trace["line"] = {
            "width": line_width,
            "color": color,
            "shape": line_shape,
            "dash": "dashdot",
        }
        traces.append(high_trace)

    # Maximum
    if data.maximum is not None:
        maximum_trace = get_default_trace(
            statistics_name="Maximum",
            values=data.maximum,
        )
        maximum_trace["line"] = {
            "color": color,
            "shape": line_shape,
            "dash": "longdash",
            "width": line_width,
        }
        traces.append(maximum_trace)

    # Free line
    if data.free_line is not None:
        line_trace = get_default_trace(
            statistics_name=data.free_line.name,
            values=data.free_line.data,
        )
        # Set solid line
        line_trace["line"] = {"width": line_width, "color": color, "shape": line_shape}
        traces.append(line_trace)

    # Set legend for last trace in list
    if len(traces) > 0:
        traces[-1]["showlegend"] = show_legend

    return traces



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

import geojson
import numpy as np

from webviz_subsurface._providers import (
    EnsembleSurfaceProvider,
    SimulatedSurfaceAddress,
    StatisticalSurfaceAddress,
    SurfaceAddress,
    SurfaceImageMeta,
    SurfaceImageServer,
)
from webviz_subsurface._providers.ensemble_surface_provider.ensemble_surface_provider import (
    SurfaceStatistic,
)
from webviz_subsurface._utils.webvizstore_functions import read_csv
from webviz_subsurface.plugins._co2_leakage._utilities import plume_extent
from webviz_subsurface.plugins._co2_leakage._utilities.generic import MapAttribute
from webviz_subsurface.plugins._co2_leakage._utilities.surface_publishing import (
    TruncatedSurfaceAddress,
    publish_and_get_surface_metadata,
)
from webviz_subsurface.plugins._map_viewer_fmu._tmp_well_pick_provider import (
    WellPickProvider,
)


def property_origin(
    attribute: MapAttribute, map_attribute_names: Dict[MapAttribute, str]
) -> str:
    if attribute in map_attribute_names:
        return map_attribute_names[attribute]
    if attribute == MapAttribute.SGAS_PLUME:
        return map_attribute_names[MapAttribute.MAX_SGAS]
    if attribute == MapAttribute.AMFG_PLUME:
        return map_attribute_names[MapAttribute.MAX_AMFG]
    raise AssertionError(f"No origin defined for property: {attribute}")


@dataclass
class SurfaceData:
    readable_name: str
    color_map_range: Tuple[Optional[float], Optional[float]]
    color_map_name: str
    value_range: Tuple[float, float]
    meta_data: SurfaceImageMeta
    img_url: str

    @staticmethod
    def from_server(
        server: SurfaceImageServer,
        provider: EnsembleSurfaceProvider,
        address: Union[SurfaceAddress, TruncatedSurfaceAddress],
        color_map_range: Tuple[Optional[float], Optional[float]],
        color_map_name: str,
        readable_name_: str,
    ) -> "SurfaceData":
        surf_meta, img_url = publish_and_get_surface_metadata(server, provider, address)
        assert surf_meta is not None  # Should not occur
        value_range = (
            0.0 if np.ma.is_masked(surf_meta.val_min) else surf_meta.val_min,
            0.0 if np.ma.is_masked(surf_meta.val_max) else surf_meta.val_max,
        )
        color_map_range = (
            value_range[0] if color_map_range[0] is None else color_map_range[0],
            value_range[1] if color_map_range[1] is None else color_map_range[1],
        )
        return SurfaceData(
            readable_name_,
            color_map_range,
            color_map_name,
            value_range,
            surf_meta,
            img_url,
        )


def derive_surface_address(
    surface_name: str,
    attribute: MapAttribute,
    date: Optional[str],
    realization: List[int],
    map_attribute_names: Dict[MapAttribute, str],
    statistic: str,
    contour_data: Optional[Dict[str, Any]],
) -> Union[SurfaceAddress, TruncatedSurfaceAddress]:
    if attribute in (MapAttribute.SGAS_PLUME, MapAttribute.AMFG_PLUME):
        assert date is not None
        basis = (
            MapAttribute.MAX_SGAS
            if attribute == MapAttribute.SGAS_PLUME
            else MapAttribute.MAX_AMFG
        )
        return TruncatedSurfaceAddress(
            name=surface_name,
            datestr=date,
            realizations=realization,
            basis_attribute=map_attribute_names[basis],
            threshold=contour_data["threshold"] if contour_data else 0.0,
            smoothing=contour_data["smoothing"] if contour_data else 0.0,
        )
    date = None if attribute == MapAttribute.MIGRATION_TIME else date
    if len(realization) == 1:
        return SimulatedSurfaceAddress(
            attribute=map_attribute_names[attribute],
            name=surface_name,
            datestr=date,
            realization=realization[0],
        )
    return StatisticalSurfaceAddress(
        attribute=map_attribute_names[attribute],
        name=surface_name,
        datestr=date,
        statistic=SurfaceStatistic(statistic),
        realizations=realization,
    )


def readable_name(attribute: MapAttribute) -> str:
    unit = ""
    if attribute == MapAttribute.MIGRATION_TIME:
        unit = " [year]"
    elif attribute in (MapAttribute.AMFG_PLUME, MapAttribute.SGAS_PLUME):
        unit = " [# real.]"
    return f"{attribute.value}{unit}"


def get_plume_polygon(
    surface_provider: EnsembleSurfaceProvider,
    realizations: List[int],
    surface_name: str,
    datestr: str,
    contour_data: Dict[str, Any],
) -> Optional[geojson.FeatureCollection]:
    surface_attribute = contour_data["property"]
    threshold = contour_data["threshold"]
    smoothing = contour_data["smoothing"]
    if (
        surface_attribute is None
        or len(realizations) == 0
        or threshold is None
        or threshold <= 0
    ):
        return None
    surfaces = [
        surface_provider.get_surface(
            SimulatedSurfaceAddress(
                attribute=surface_attribute,
                name=surface_name,
                datestr=datestr,
                realization=r,
            )
        )
        for r in realizations
    ]
    surfaces = [s for s in surfaces if s is not None]
    if len(surfaces) == 0:
        return None
    return plume_extent.plume_polygons(
        surfaces,
        threshold,
        smoothing=smoothing,
        simplify_factor=0.12 * smoothing,  # Experimental
    )


def create_map_layers(
    formation: str,
    surface_data: Optional[SurfaceData],
    fault_polygon_url: Optional[str],
    license_boundary_file: Optional[str],
    well_pick_provider: Optional[WellPickProvider],
    plume_extent_data: Optional[geojson.FeatureCollection],
) -> List[Dict]:
    layers = []
    if surface_data is not None:
        # Update ColormapLayer
        layers.append(
            {
                "@@type": "ColormapLayer",
                "name": surface_data.readable_name,
                "id": "colormap-layer",
                "image": surface_data.img_url,
                "bounds": surface_data.meta_data.deckgl_bounds,
                "valueRange": surface_data.value_range,
                "colorMapRange": surface_data.color_map_range,
                "colorMapName": surface_data.color_map_name,
                "rotDeg": surface_data.meta_data.deckgl_rot_deg,
            }
        )

    if fault_polygon_url is not None:
        layers.append(
            {
                "@@type": "FaultPolygonsLayer",
                "name": "Fault Polygons",
                "id": "fault-polygons-layer",
                "data": fault_polygon_url,
            }
        )
    if license_boundary_file is not None:
        layers.append(
            {
                "@@type": "FaultPolygonsLayer",
                "name": "Containment Boundary",
                "id": "license-boundary-layer",
                "data": _parse_polygon_file(license_boundary_file),
            }
        )
    if well_pick_provider is not None:
        layers.append(
            {
                "@@type": "GeoJsonLayer",
                "name": "Well Picks",
                "id": "well-picks-layer",
                "data": dict(
                    well_pick_provider.get_geojson(
                        well_pick_provider.well_names(), formation
                    )
                ),
            }
        )
    if plume_extent_data is not None:
        layers.append(
            {
                "@@type": "GeoJsonLayer",
                "name": "Plume Contours",
                "id": "plume-polygon-layer",
                "data": dict(plume_extent_data),
                "lineWidthMinPixels": 2,
                "getLineColor": [150, 150, 150, 255],
            }
        )
    return layers


def _parse_polygon_file(filename: str) -> Dict[str, Any]:
    xyz = read_csv(filename)[["x", "y"]].values
    as_geojson = {
        "type": "FeatureCollection",
        "features": [
            {
                "type": "Feature",
                "properties": {},
                "geometry": {
                    "type": "LineString",
                    "coordinates": xyz.tolist(),
                },
            }
        ],
    }
    return as_geojson



import io
import warnings
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import xtgeo
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore


class SurfaceSetModel:
    """Class to load and calculate statistical surfaces from an FMU Ensemble"""

    def __init__(self, surface_table: pd.DataFrame):
        self._surface_table = surface_table

    @property
    def realizations(self) -> list:
        """Returns surface attributes"""
        return sorted(list(self._surface_table["REAL"].unique()))

    @property
    def attributes(self) -> list:
        """Returns surface attributes"""
        return list(self._surface_table["attribute"].unique())

    def names_in_attribute(self, attribute: str) -> list:
        """Returns surface names for a given attribute"""
        return list(
            self._surface_table.loc[self._surface_table["attribute"] == attribute][
                "name"
            ].unique()
        )

    def dates_in_attribute(self, attribute: str) -> list:
        """Returns surface dates for a given attribute"""
        return list(
            self._surface_table.loc[self._surface_table["attribute"] == attribute][
                "date"
            ].unique()
        )

    def get_realization_surfaces(
        self,
        name: str,
        attribute: str,
        realizations: List[int],
        date: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        return [
            {
                "REAL": realization,
                "surface": self.get_realization_surface(
                    name=name,
                    attribute=attribute,
                    realization=int(realization),
                    date=date,
                ),
            }
            for realization in realizations
        ]

    def get_realization_surface(
        self, name: str, attribute: str, realization: int, date: Optional[str] = None
    ) -> xtgeo.RegularSurface:
        """Returns a Xtgeo surface instance of a single realization surface"""

        columns = ["name", "attribute", "REAL"]
        column_values = [name, attribute, realization]
        if date is not None:
            columns.append("date")
            column_values.append(date)

        df = self._filter_surface_table(
            name=name, attribute=attribute, date=date, realizations=[int(realization)]
        )
        if len(df.index) == 0:
            warnings.warn(
                f"No surface found for name: {name}, attribute: {attribute}, date: {date}, "
                f"realization: {realization}"
            )
            return xtgeo.RegularSurface(
                ncol=1, nrow=1, xinc=1, yinc=1
            )  # 1's as input is required
        if len(df.index) > 1:
            warnings.warn(
                f"Multiple surfaces found for name: {name}, attribute: {attribute}, date: {date}, "
                f"realization: {realization}. Returning first surface"
            )
        return xtgeo.surface_from_file(get_stored_surface_path(df.iloc[0]["path"]))

    def _filter_surface_table(
        self,
        name: str,
        attribute: str,
        date: Optional[str] = None,
        realizations: Optional[List[int]] = None,
    ) -> pd.DataFrame:
        """Returns a dataframe of surfaces for the provided filters"""
        columns: List[str] = ["name", "attribute"]
        column_values: List[Any] = [name, attribute]
        if date is not None:
            columns.append("date")
            column_values.append(date)
        if realizations is not None:
            columns.append("REAL")
            column_values.append(realizations)
        df = self._surface_table.copy()
        for filt, col in zip(column_values, columns):
            if isinstance(filt, list):
                df = df.loc[df[col].isin(filt)]
            else:
                df = df.loc[df[col] == filt]
        return df

    @CACHE.memoize(timeout=CACHE.TIMEOUT)
    def calculate_statistical_surface(
        self,
        name: str,
        attribute: str,
        calculation: Optional[str] = "Mean",
        date: Optional[str] = None,
        realizations: Optional[List[int]] = None,
    ) -> xtgeo.RegularSurface:
        """Returns a Xtgeo surface instance for a calculated surface"""
        df = self._filter_surface_table(
            name=name, attribute=attribute, date=date, realizations=realizations
        )
        # When portable check if the surface has been stored
        # if not calculate
        try:
            surface_stream = save_statistical_surface(
                sorted(list(df["path"])), calculation
            )
        except OSError:
            surface_stream = save_statistical_surface_no_store(
                sorted(list(df["path"])), calculation
            )

        return xtgeo.surface_from_file(surface_stream, fformat="irap_binary")

    def webviz_store_statistical_calculation(
        self,
        calculation: Optional[str] = "Mean",
        realizations: Optional[List[int]] = None,
    ) -> Tuple[Callable, list]:
        """Returns a tuple of functions to calculate statistical surfaces for
        webviz store"""
        df = (
            self._surface_table.loc[self._surface_table["REAL"].isin(realizations)]
            if realizations is not None
            else self._surface_table
        )
        stored_functions_args = []
        for _attr, attr_df in df.groupby("attribute"):
            for _name, name_df in attr_df.groupby("name"):

                if name_df["date"].isnull().values.all():
                    stored_functions_args.append(
                        {
                            "fns": sorted(list(name_df["path"].unique())),
                            "calculation": calculation,
                        }
                    )
                else:
                    for _date, date_df in name_df.groupby("date"):
                        stored_functions_args.append(
                            {
                                "fns": sorted(list(date_df["path"].unique())),
                                "calculation": calculation,
                            }
                        )

        return (
            save_statistical_surface,
            stored_functions_args,
        )

    def webviz_store_realization_surfaces(self) -> Tuple[Callable, list]:
        """Returns a tuple of functions to store all realization surfaces for
        webviz store"""
        return (
            get_stored_surface_path,
            [{"runpath": path} for path in list(self._surface_table["path"])],
        )

    @property
    def first_surface_geometry(self) -> Dict:
        surface = xtgeo.surface_from_file(
            get_stored_surface_path(self._surface_table.iloc[0]["path"])
        )
        return {
            "xmin": surface.xmin,
            "xmax": surface.xmax,
            "ymin": surface.ymin,
            "ymax": surface.ymax,
            "xori": surface.xori,
            "yori": surface.yori,
            "ncol": surface.ncol,
            "nrow": surface.nrow,
            "xinc": surface.xinc,
            "yinc": surface.yinc,
        }


@webvizstore
def get_stored_surface_path(runpath: Path) -> Path:
    """Returns path of a stored surface"""
    return Path(runpath)


def save_statistical_surface_no_store(
    fns: List[str], calculation: Optional[str] = "Mean"
) -> io.BytesIO:
    """Wrapper function to store a calculated surface as BytesIO"""

    surfaces = xtgeo.Surfaces([get_stored_surface_path(fn) for fn in fns])
    if len(surfaces.surfaces) == 0:
        surface = xtgeo.RegularSurface(
            ncol=1, nrow=1, xinc=1, yinc=1
        )  # 1's as input is required
    elif calculation in ["Mean", "StdDev", "Min", "Max", "P10", "P90"]:
        # Suppress numpy warnings when surfaces have undefined z-values
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "All-NaN slice encountered")
            warnings.filterwarnings("ignore", "Mean of empty slice")
            warnings.filterwarnings("ignore", "Degrees of freedom <= 0 for slice")
            surface = get_statistical_surface(surfaces, calculation)
    else:
        surface = xtgeo.RegularSurface(
            ncol=1, nrow=1, xinc=1, yinc=1
        )  # 1's as input is required
    stream = io.BytesIO()
    surface.to_file(stream, fformat="irap_binary")
    return stream


@webvizstore
def save_statistical_surface(fns: List[str], calculation: str) -> io.BytesIO:
    """Wrapper function to store a calculated surface as BytesIO"""
    surfaces = xtgeo.Surfaces(fns)
    if len(surfaces.surfaces) == 0:
        surface = xtgeo.RegularSurface(
            ncol=1, nrow=1, xinc=1, yinc=1
        )  # 1's as input is required
    elif calculation in ["Mean", "StdDev", "Min", "Max", "P10", "P90"]:
        # Suppress numpy warnings when surfaces have undefined z-values
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "All-NaN slice encountered")
            warnings.filterwarnings("ignore", "Mean of empty slice")
            warnings.filterwarnings("ignore", "Degrees of freedom <= 0 for slice")
            surface = get_statistical_surface(surfaces, calculation)
    else:
        surface = xtgeo.RegularSurface(
            ncol=1, nrow=1, xinc=1, yinc=1
        )  # 1's as input is required
    stream = io.BytesIO()
    surface.to_file(stream, fformat="irap_binary")
    return stream


# pylint: disable=too-many-return-statements
def get_statistical_surface(
    surfaces: xtgeo.Surfaces, calculation: str
) -> xtgeo.RegularSurface:
    """Calculates a statistical surface from a list of Xtgeo surface instances"""
    if calculation == "Mean":
        return surfaces.apply(np.mean, axis=0)
    if calculation == "StdDev":
        return surfaces.apply(np.std, axis=0)
    if calculation == "Min":
        return surfaces.apply(np.min, axis=0)
    if calculation == "Max":
        return surfaces.apply(np.max, axis=0)
    if calculation == "P10":
        return surfaces.apply(np.percentile, 10, axis=0)
    if calculation == "P90":
        return surfaces.apply(np.percentile, 90, axis=0)
    return xtgeo.RegularSurface(
        ncol=1, nrow=1, xinc=1, yinc=1
    )  # 1's as input is required



from .comparison_controllers import comparison_controllers
from .distribution_controllers import distribution_controllers
from .export_data_controllers import export_data_controllers
from .fipfile_qc_controller import fipfile_qc_controller
from .layout_controllers import layout_controllers
from .selections_controllers import selections_controllers
from .tornado_controllers import tornado_controllers



from typing import List

import pandas as pd


class CorrelationFigure:
    def __init__(self, series: pd.Series, n_rows: int, title: str):
        self.series = series.tail(n=n_rows)
        self.title = title
        self._data = self.initial_data

    @property
    def data(self) -> List[dict]:
        return self._data

    @property
    def figure(self) -> dict:
        return {"data": self.data, "layout": self.layout}

    @property
    def initial_data(self) -> List[dict]:
        return [
            {
                "x": self.series.values,
                "y": self.series.index,
                "orientation": "h",
                "type": "bar",
                "marker": {},
            }
        ]

    @property
    def layout(self) -> dict:
        return {
            "barmode": "relative",
            "xaxis": {"range": [-1, 1]},
            "yaxis": {"automargin": True},
            "automargin": True,
        }

    @property
    def first_y_value(self) -> str:
        return self.data[0]["y"][-1]

    def set_bar_colors(self, selected_bar: str) -> None:
        self._data[0]["marker"]["color"] = [
            "grey" if _bar != selected_bar else "red" for _bar in self.data[0]["y"]
        ]



from .ensemble_model import EnsembleModel
from .ensemble_set_model import EnsembleSetModel
from .gruptree_model import GruptreeModel
from .inplace_volumes_model import InplaceVolumesModel
from .observation_model import ObservationModel
from .parameter_model import ParametersModel
from .stratigraphy_model import StratigraphyModel
from .surface_leaflet_model import SurfaceLeafletModel
from .surface_set_model import SurfaceSetModel
from .well_attributes_model import WellAttributesModel
from .well_set_model import WellSetModel



import datetime
import fnmatch
import re
from typing import List, Optional, Set

import pandas as pd

from webviz_subsurface._abbreviations.reservoir_simulation import historical_vector
from webviz_subsurface._providers import EnsembleSummaryProvider
from webviz_subsurface._utils.simulation_timeseries import (
    set_simulation_line_shape_fallback,
)
from webviz_subsurface._utils.vector_selector import add_vector_to_vector_selector_data


class ProviderTimeSeriesDataModel:
    """Class to process and and visualize ensemble timeseries"""

    def __init__(self, provider_set: dict, column_keys: Optional[list] = None) -> None:

        self._provider_set = provider_set
        self.line_shape_fallback = set_simulation_line_shape_fallback("linear")
        all_vector_names = self._create_union_of_vector_names_from_providers(
            list(provider_set.values())
        )
        self._vector_names = (
            self.filter_vectorlist_on_column_keys(column_keys, all_vector_names)
            if column_keys is not None
            else all_vector_names
        )
        if not self._vector_names:
            raise ValueError("No vectors match the selected 'column_keys' criteria")

        self._dates = self.all_dates()

        # add vectors to vector selector
        self.vector_selector_data: list = []
        for vector in self.get_non_historical_vector_names():
            add_vector_to_vector_selector_data(self.vector_selector_data, vector)

    @property
    def vectors(self) -> List[str]:
        return self._vector_names

    @property
    def dates(self) -> List[datetime.datetime]:
        return self._dates

    def get_non_historical_vector_names(self) -> list:
        return [
            vector
            for vector in self._vector_names
            if historical_vector(vector, None, False) not in self._vector_names
        ]

    def all_dates(self) -> List[datetime.datetime]:
        """List with the union of dates among providers"""
        # TODO: Adjust when providers are updated!
        dates_union: Set[datetime.datetime] = set()
        for provider in list(self._provider_set.values()):
            _dates = set(provider.dates(None))
            dates_union.update(_dates)
        return list(sorted(dates_union))

    @staticmethod
    def _create_union_of_vector_names_from_providers(
        providers: List[EnsembleSummaryProvider],
    ) -> List[str]:
        """Create list with the union of vector names among providers"""
        vector_names = []
        for provider in providers:
            vector_names.extend(
                provider.vector_names_filtered_by_value(
                    exclude_all_values_zero=True, exclude_constant_values=True
                )
            )
        vector_names = list(sorted(set(vector_names)))
        return vector_names

    @staticmethod
    def filter_vectorlist_on_column_keys(
        column_key_list: list, vectorlist: list
    ) -> list:
        """Filter vectors using list of unix shell wildcards"""
        try:
            regex = re.compile(
                "|".join([fnmatch.translate(col) for col in column_key_list]),
                flags=re.IGNORECASE,
            )
            return [v for v in vectorlist if regex.fullmatch(v)]
        except re.error:
            return []

    def get_historical_vector_df(
        self, vector: str, ensemble: str
    ) -> Optional[pd.DataFrame]:
        hist_vecname = historical_vector(vector, smry_meta=None)

        if hist_vecname and hist_vecname in self.vectors:
            provider = self._provider_set[ensemble]
            return provider.get_vectors_df(
                [hist_vecname], None, realizations=provider.realizations()[:1]
            ).rename(columns={hist_vecname: vector})
        return None

    def get_vector_df(
        self,
        ensemble: str,
        realizations: List[int],
        vectors: List[str],
    ) -> pd.DataFrame:
        provider = self._provider_set[ensemble]
        ens_vectors = [vec for vec in vectors if vec in provider.vector_names()]
        return provider.get_vectors_df(ens_vectors, None, realizations)

    def get_last_date(self, ensemble: str) -> str:
        return max(self._provider_set[ensemble].dates(None))



import datetime
import hashlib
import logging
import os
import pickle  # nosec
import uuid
from pathlib import Path
from typing import Optional

import xtgeo

from .ensemble_surface_provider import StatisticalSurfaceAddress

LOGGER = logging.getLogger(__name__)

# For some obscure reason, reading of a non-existent irap file segfaults,
# so use asymmetric file formats for read and write
FILE_FORMAT_WRITE = "irap_binary"
FILE_FORMAT_READ = "guess"
FILE_EXTENSION = ".gri"

# FILE_FORMAT_WRITE = "xtgregsurf"
# FILE_FORMAT_READ = FILE_FORMAT_WRITE
# FILE_EXTENSION = ".xtgregsurf"


class StatSurfCache:
    def __init__(self, cache_dir: Path) -> None:
        self.cache_dir = cache_dir

        self.cache_dir.mkdir(parents=True, exist_ok=True)
        placeholder_file = self.cache_dir / "placeholder.txt"
        placeholder_file.write_text(
            f"Placeholder -- {datetime.datetime.now()} -- {os.getpid()}"
        )

    def fetch(
        self, address: StatisticalSurfaceAddress
    ) -> Optional[xtgeo.RegularSurface]:

        full_surf_path = self.cache_dir / _compose_stat_surf_file_name(
            address, FILE_EXTENSION
        )

        try:
            surf = xtgeo.surface_from_file(full_surf_path, fformat=FILE_FORMAT_READ)
            return surf
        # pylint: disable=bare-except
        except:
            return None

    def store(
        self, address: StatisticalSurfaceAddress, surface: xtgeo.RegularSurface
    ) -> None:

        surf_fn = _compose_stat_surf_file_name(address, FILE_EXTENSION)
        full_surf_path = self.cache_dir / surf_fn

        # Try and go via a temporary file which we don't rename until writing is finished.
        # to make the cache writing more concurrency-friendly.
        # One problem here is that we don't control the file handle (xtgeo does) so can't
        # enforce flush and sync of the file to disk before the rename :-(
        # Still, we probably need a more robust way of shring the cached surfaces...
        tmp_surf_path = self.cache_dir / (surf_fn + f"__{uuid.uuid4().hex}.tmp")
        try:
            surface.to_file(tmp_surf_path, fformat=FILE_FORMAT_WRITE)
            os.replace(tmp_surf_path, full_surf_path)
        # pylint: disable=bare-except
        except:
            os.remove(tmp_surf_path)

        # surface.to_file(full_surf_path, fformat=FILE_FORMAT_WRITE)


def _compose_stat_surf_file_name(
    address: StatisticalSurfaceAddress, extension: str
) -> str:

    # Should probably sort the realization list
    # Also, what about duplicates
    # And further, handling of missing realizations...

    pickled = pickle.dumps(address.realizations, pickle.HIGHEST_PROTOCOL)
    real_hash = hashlib.md5(pickled).hexdigest()  # nosec
    return "--".join(
        [
            f"{address.statistic}",
            f"{address.name}",
            f"{address.attribute}",
            f"{address.datestr}",
            f"{real_hash}{extension}",
        ]
    )
